/*-- Copyright (c) 2012 by SAS Institute Inc., Cary, NC, USA. --*/ 
/*-- All Rights Reserved.                                     --*/
 
/* ****************************************************************************
Macro         PAM_UPDATE_DATA_SEL_TABLE.SAS

Instructions  Run the pam_update_data_sel macro on the PAM 5.2 server.

Purpose       Update the DATA_SEL table in the Domain Object library to 
              increase the length of columns FILTERED_SUBSET_ROWS_NO and
              FILTERED_SUBSET_SIZE_NO.

Macro Name    pam_update_data_sel_table

****************************************************************************** */

%pam_prebuild;

%macro pam_update_data_sel_table;
%let connectionString=;
data _null_;
  length libref $8 libraryOptions $1024 uri uri2 $256 ;

  call missing(libref, libraryOptions, uri, uri2);
  n=1;
  n2=1;

  /* Go through the Library objects looking for the domain object library. */
  rc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",n,uri);
  if rc<=0 then put "NOTE: rc=" rc 
    "There are no Libraries defined in this repository"
    " or there was an error reading the repository.";

  do while(rc>0);
    objrc=metadata_getattr(uri,"Libref",libref);

    /* Get the Properties association for the domain object library. */
    if upcase(libref) eq upcase("&pam_dolibref") then do;
      objrc=metadata_getnasn(uri,"Properties",n2,uri2);
      if objrc>=0 then do;
        objrc=metadata_getattr(uri2,"DefaultValue",libraryOptions);
        call symput('connectionString',libraryOptions);
      end;
    end;

    n+1;

    /* Continue looking for the domain object library if not found yet. */
    n2=1;

    rc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",n,uri);
  end; /*do while rc*/  
run;

/* Perform the update. */
proc sql;
  connect to sasts(&connectionString);
  execute(alter table DATA_SEL alter filtered_subset_rows_no type double precision) by sasts;
  execute(alter table DATA_SEL alter filtered_subset_size_no type numeric(18)) by sasts;
quit;

%mend;
%pam_update_data_sel_table;
