/*-- Copyright (c) 2012 by SAS Institute Inc., Cary, NC, USA. --*/ 
/*-- All Rights Reserved.                                     --*/
 
/* ****************************************************************************
Macro         PAM_EXPAND_TRANSPOSE.SAS
Purpose       Take 1 or 2 input data sets (a table with tags/measures and/or a 
              table with events)and expand the specified datetime variable 
              values into a complete time seres within the beginning and ending 
              datetime specified. The values of the specified variable will be 
              transposed by the group variables and datetime variable. The 
              values will be interpolated and/or aggregated as specified.   
Macro Name    pam_expand_transpose
 
History       12/15/2011 Version 5.1_1
              08/17/2012 Version 5.2_1 - PAM 5.2 - support for events, 
                                         interpolation by tag, and aggregation

Parameters:

outds            = Required output data set to hold the resulting records.
groupvar         = List of by variables that will be used to group the 
                   transposed records.
                   Default: pam_asset_dim_rk pam_loc_dim_rk
periodicity      = Number of seconds for fixed periodicity. If blank or 0, 
                   then no periodicity will be applied. Must be a positive 
                   interger.
                   Default: blank
begindttm        = Required datetime value.  The resulting time series in 
                   OUTDS will begin with this value.
enddttm          = Required datetime value.  The resulting time series in 
                   OUTDS will end with this value.
timeslicebegin   = Optional time, when used with &timeslicend, restricts the 
                   expanded observations to be within the beginning and ending 
                   times specified, for all of the dates occuring within the 
                   ranges specified by the &begindttm and &enddttm values.
timesliceend     = Optional time, used with &timeslicebegin.
measure_inds     = input table containing tags/measures.  
                   Either this parm or the &event_inds parm is required.
measure_dttmvar  = Datetime variable name from the &measure_inds data set. 
                   Default: measure_Dttm
measure_rkvar    = &measure_inds variable whose values are the basis for 
                   transposed columns (RK variable). 
                   Default: pam_tag_dim_rk 
measure_var      = &measure_inds variable whose values will be transposed.
                   Default: loc_measure_value 
measure_dim      = Name of the dimension table for tags.  
                   Default: &pam_dimLibref..pam_tag_dim
measure_colnamevar  = Column Name variable in the &measure_dim.  
                   Default: tag_column_nm
measure_intvar   = Interpolation Method variable in the &measure_dim.
                   Default: tag_def_interpol_method_cd
measure_aggvar   = Aggregation Method variable in the &measure_dim.
                   Default: tag_def_aggreg_method_cd
measure_int      = Interpolation method that the user specified in the UI for 
                   measure data. "DEFAULT" or blank means to use the method 
                   from the &measure_dim table for each transposed column.
measure_agg      = Aggregation method that the user specified in the UI for 
                   measure data. "DEFAULT" or blank means to use the method 
                   from the &measure_dim table for each transposed column. 
measure_rklist   = Name of the table that contains the list of tag RKs whose 
                   values will be used to create the tansposed columns.
                   Required if &measure_inds is not blank.
measure_idvar    = &measure_rklist variable that contains the ID variable. 
                   Default: tag_id 
measure_filter   = Optional data filter for measure data.  If &measure_filterRKs
                   is populated, then &measure_filter must also be populated. 
measure_filterRKs = Name of the table that contains the tag RKs that have filters.
                   If &measure_filter is populated, then &measure_filterRKs must 
                   also be populated.
measure_RKsNoFilter = Name of the table that contains the tag RKs that do not 
                   have filters.
                   If &measure_filter is populated, then &measure_filterRKs must 
                   also be populated.
event_inds       = input table containing events.  
                   Either this parm or the &measure_inds parm is required.
event_dttmvar    = Datetime variable name from the &event_inds data set. 
                   Default: event_dttm
event_rkvar      = event_inds variable whose values are the basis for transposed 
                   columns (RK variable). 
                   Default: pam_event_dim_rk 
event_var        = &event_inds variable whose values will be transposed.
                   Default: event_value
event_dim        = Name of the dimension table for events.  
                   Default: &pam_dimLibref..pam_event_dim
event_colnamevar = Column Name variable in the &event_dim.  
                   Default: event_column_nm
event_aggvar     = Aggregation Method variable in the &event_dim.
                   Required if &event_inds is not blank.
event_agg        = Aggregation method that the user specified in the UI for 
                   event data. "DEFAULT" or blank means to use the method from 
                   the &event_dim table for each transposed column. 
event_rklist     = Name of the table that contains the list of event RKs whose 
                   values will be used to create the tansposed columns.
                   Required if &event_inds is not blank.
event_idvar      = &event_rklist variable that contains the ID variable. 
                   Default: event_id 
event_filter     = Name of the table that contains the tag RKs that have filters.
                   If &event_filterRKs is populated, then &event_filter must also 
                   be populated. 
event_filterRKs  = Name of the table that contains the event RKs that do not 
                   have filters.
                   If &event_filter is populated, then &event_filterRKs must 
                   also be populated.
event_RKsNoFilter = Name of the table that contains the event RKs that do not 
                   have filters.
                   If &event_filter is populated, then &event_filterRKs must 
                   also be populated.

An example execution of this macro using as many defaults as possible:

%pam_expand_transpose( outds=outdata, measure_inds=measures, event_inds=events,
                       measure_rklist=rkList_m, event_rklist=rkList_e,
                       begindttm=20sep2011:10:00:00, 
                       enddttm=30sep2011:12:00:00 )

In the above example, the output table will contain event and measure data and 
  will be transposed by the default by variables:pam_asset_dim_rk pam_loc_dim_rk
  and datetime. It will contain only the datetime values that are in the 
  incoming data.  Missing measure values will be interpolated as specified in 
  the pam_tag_dim table. 

Another example is:

%pam_expand_transpose( outds=outdata, measure_inds=measures, event_inds=events,
                       measure_rklist=rkList_m, event_rklist=rkList_e,
                       measure_int=STEP, periodicity=60,
                       begindttm=20sep2011:10:00:00, 
                       enddttm=30sep2011:12:00:00 )

In the above example, the output table will contain event and measure data and 
  will be transposed by the default by variables:pam_asset_dim_rk pam_loc_dim_rk
  and datetime. The datetime values in the table will be aggregated or expanded 
  as appropriate to create evenly spaced data at 1 minute (60 second 
  periodicity) increments.  The measure data will be aggregated based on the 
  aggregation methods specified in the pam_tag_dim table, and will be 
  interpolated using the STEP method.  The event data will be aggregated based 
  on the aggregation methods specified in the pam_tag_dim table.  
****************************************************************************** */

%macro pam_expand_transpose(outds=,
                            groupVar=,
                            periodicity=,
                            beginDTTM=,
                            endDTTM=,
                            timeSliceBegin=,
                            timeSliceEnd=,
                            measure_inds=,
                            measure_dttmVar=,
                            measure_rkVar=,
                            measure_var=,
                            measure_dim=,
                            measure_colnameVar=,
                            measure_intVar=,
                            measure_aggVar=,
                            measure_int=,
                            measure_agg=,
                            measure_rklist=,
                            measure_idVar=,
                            measure_filter=,
                            measure_filterRKs=,
                            measure_RKsNoFilter=,
                            event_inds=,
                            event_dttmVar=,
                            event_rkVar=,
                            event_var=,
                            event_dim=,
                            event_colnameVar=,
                            event_aggVar=,
                            event_agg=,
                            event_rklist=,
                            event_idVar=,
                            event_filter=,
                            event_filterRKs=,
                            event_RKsNoFilter=);

   %put %quote(pam_expand_transpose: Version 5.2_1);

   /* set debug macro variable to 1 to avoid deleting */
   /* temporary tables and macro variables            */
   %global debug;

   /* valid aggregation values */
   %let validAggValues = 'MEAN', 'MIN', 'MAX', 'LAST', 'SUM', 'NONE';

   /* === M A C R O S === */

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_cleanup                               */
   /* Delete a table unless debug is set to 1.                          */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_cleanup(table);
   
      %global debug;
      
      %if (&debug ne 1) %then 
      %do; 
      
         %if (%sysfunc(exist(&table))) %then 
         %do;
            proc delete data = &table;
            run;
         %end;
         
      %end; /* if - debug is on */
      
   %mend pam_expand_transpose_cleanup;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_checkInds                             */
   /* check that the type_inds table exists                             */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_checkInds(type);

      %if (%nrquote(&&&type._inds) ne ) %then 
      %do;

         %if (not %sysfunc(exist(&&&type._inds))) %then 
            %let &type._inds = ;

      %end;

   %mend pam_expand_transpose_checkInds;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_setDefault                            */
   /* Set the default value for a blank parameter                       */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_setDefault(parm, default);
   
      %if (%nrquote(&&&parm) eq ) %then 
         %let &parm = &default;
         
   %mend pam_expand_transpose_setDefault;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_nonBlank                              */
   /* Put out an error if a parameter has a blank value                 */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_nonBlank(parm);
   
      %if (%nrquote(&&&parm) eq ) %then 
      %do; 
         %put ERROR%quote(:) &primaryMacro requires a value for parameter &parm;
         %abort cancel 8;
      %end;
      
   %mend pam_expand_transpose_nonBlank;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_dsExist                               */
   /* Put out an error if a required data table does not exist          */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_dsExist(parm);
   
      %if (not %sysfunc(exist(%nrquote(&&&parm)))) %then 
      %do; 
         %let msg = &primaryMacro requires an existing data set for;
         %let msg = &msg parameter &parm;
         %put ERROR%quote(:) &msg;
         %abort cancel 8;
      %end;
      
   %mend pam_expand_transpose_dsExist;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_varExist                              */
   /* Put out an error if a variable does not exist in a data table     */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_varExist(dsid, parm, msg);
   
      %local varnum;
      %let varnum = 0;
      %if (&dsid gt 0) %then 
         %let varnum = %sysfunc(varnum(&dsid, &&&parm));

      %if (&varnum le 0) %then 
      %do;
      
         %if (%nrquote(&msg) eq ) %then 
         %do;
            %let msg = &primaryMacro requires an existing variable;
            %let msg = &msg for the parameter &parm;
         %end;
         
         %put ERROR%quote(:) %nrquote(&msg);

         %let dsid = %sysfunc(close(&dsid));
         %abort cancel 8;
         
      %end; /* if - variable not found */
      
   %mend pam_expand_transpose_varExist;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_filterRK                              */
   /* Make macro variables from values in type_filterRKs and            */
   /* type_RKsNoFilter tables.                                          */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_filterRK(type);

      %global &type._numFilter &type._numNoFilter;
      %let &type._numFilter = 0;
      %let &type._numNoFilter = 0;

      /* only need to create these macro variables if there is a filter */
      %if (%superq(&type._filter) eq ) %then 
         %return;

      data _null_;
         set &&&type._filterRKs end = eof;
         call symputx(cats("&type._filterRK", _n_), &&&type._rkVar, "G");
         if (eof) then 
            call symputx("&type._numFilter", _n_);
      run;

      data _null_;
         set &&&type._RKsNoFilter end = eof;
         call symputx(cats("&type._noFilterRK", _n_), &&&type._rkVar, "G");
         if (eof) then 
            call symputx("&type._numNoFilter", _n_);
      run;

   %mend pam_expand_transpose_filterRK;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_validAgg                              */
   /* Put out an error if the measure or event agg parm does not have a */
   /* valid value (MEAN, MIN MAX, LAST, SUM, DEFAULT)                   */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_validAgg(agg);

      %let &agg = %kupcase(%nrquote(&&&agg));

      data _null_;
         agg = symget("&agg");
         if (agg not in (&validAggValues, "DEFAULT", "")) then 
         do; 
            length msg $200;
            msg = "&primaryMacro requires a valid value for &agg.";
            put "ERR" "OR: " msg;
            msg = "Valid values: &validAggValues, 'DEFAULT', blank";
            put "ERR" "OR: " msg;
            abort cancel 8;
         end;
      run;

      %if (%nrquote(&&&agg) eq DEFAULT) %then 
         %let &agg = ;
         
   %mend pam_expand_transpose_validAgg;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_getColInfo                            */
   /* Get all the information needed for the measure and event columns  */
   /* that will be created in the final table                           */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_getColInfo(type);

      /* don't do anything if there is no data for this type */
      %if (not &&&type) %then 
         %return;

      /* put together the data from the list of valid ids and the dim table */
      proc sql;
         create table rklist as 
            select i.&&&type._rkVar as rk,
                   d.&&&type._idVar as id,
                   d.&&&type._colnameVar as colname,
                   upcase(coalesce("&&&type._agg", d.&&&type._aggVar)) as agg,
                   %if (%upcase(&type) eq MEASURE) %then 
                   %do;
                      %do g = 1 %to &numGroup;
                         d.&&groupVar&g,
                      %end;                   
                      upcase(coalesce("&&&type._int", d.&&&type._intVar)) as int
                   %end;
                   %else
                   %do;
                      " " as int
                   %end;
              from &&&type._rklist as i
                      left join
                   &&&type._dim as d
                 on i.&&&type._rkVar eq d.&&&type._rkVar and
                    d.current_num eq 1;
      quit;

      /* throw out any duplicate RKs */
      proc sort data = rklist nodupkey;
         by rk;
      run;

      /* make format of RK->ID */
      data fmtdata(keep=start label fmtname type);
         set rklist(rename=(rk=start 
                            id=label));
         retain fmtname "&type.RKfmt" type "N";
      run;
      proc format cntlin=fmtdata;
      run;
      %pam_expand_transpose_cleanup(fmtdata);

      /* sort data so so that we can see if we have duplicate column names */
      %if (%upcase(&type) eq MEASURE) %then 
      %do; 
         %let byvar = &groupVar;
      %end;
      %else 
      %do; 
         %let byvar = ;
      %end;

      proc sort data = rklist;
         by &byvar colname;
      run;

      data uniqueNames (keep = colname);
         set rklist end=eof;
         by &byvar colname;

         /* validate the column name to be used for transposing */
         if (not nvalid(colname, 'V7')) then 
         do; 
            length msg $200;
            msg = 'ERR' !! 'OR: ' !! "For &type id, " !! ktrim(kleft(id)) !!
                  ", the transposed column name '" !! ktrim(kleft(colname)) !! 
                  "' is an invalid SAS column name.  Please contact your " !! 
                  "system adminsitrator to correct the column name in the " !!
                  "dimension table."; 
            put msg; 
            abort cancel 8;
         end;

         output;

         /* transpose column name */
         call symputx(cats("&type._name", _n_), colname, "G");
         /* rk value */
         call symputx(cats("&type._rk", _n_), rk, "G");
         /* id value */
         call symputx(cats("&type._id", _n_), id, "G");

         %if (%upcase(&type) eq EVENT) %then 
         %do; 
            /* for events, need a name for the count column */
            call symputx(cats("&type._nameN", _n_), cats(colname, "_N"), "G");
         %end;
         %else
         %do; 
            /* for measure data, need the interpolation method */
            if (int eq "") then 
               int = "NONE";

            if (int not in ("NONE", "STEP")) then 
            do; 
               length msg $200;
               msg = "The interpolation method, '" !! 
                     strip(int) !! "', is invalid for measure id, " !! strip(id);
               put "ERR" "OR: " msg;
               msg = "Valid values are: 'STEP', 'NONE'";
               put "ERR" "OR: " msg;
               abort cancel 8;
            end;
               
            call symputx(cats("&type._int", _n_), upcase(int), "G");

            /* keep the group values */
            %do g = 1 %to &numGroup;
               call symputx(cats("group&g", _n_), &&groupVar&g, "G");
            %end;
         %end;

         /* aggregation method */
         if (agg in ("", "DEFAULT")) then 
            agg = "NONE";
         else
            agg = upcase(agg);
            
         if (agg eq "AVERAGE") then 
            agg = "MEAN";
            
         /* check that the agg valis is valid */
         if (agg not in (&validAggValues)) then 
         do; 
            length msg $200;
            msg = "The aggregation method, '" !! 
                  strip(agg) !! "', is invalid for &type id, " !! strip(id);
            put "ERR" "OR: " msg;
            msg = "Valid values are: &validAggValues.";
            put "ERR" "OR: " msg;
            abort cancel 8;
         end;         
         
         call symputx(cats("&type._agg", _n_), strip(agg), "G");

         /* number of columns */
         if (eof) then 
            call symputx("&type._num", _n_); 

      run;

     %pam_expand_transpose_cleanup(rklist);
         
      proc sort data = uniqueNames nodupkey;
         by colname;
      run;
      
      data _null_;
         set uniqueNames end=eof;
         call symputx(cats("&type._uniqueName", _n_), strip(colName), "G");
         %if (%upcase(&type) eq EVENT) %then 
         %do; 
            /* for events, need a name for the count column */
            call symputx(cats("&type._uniqueNameN", _n_), 
                         cats(colname, "_N"), "G");
         %end;
         if (eof) then 
            call symputx("&type._numUnique", _n_);
      run;      
    
      %pam_expand_transpose_cleanup(uniqueNames);

   %mend pam_expand_transpose_getColInfo;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_filter                                */
   /* Create the if statements from the filter parameters for either    */
   /* measure data or event data                                        */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_filter(type);

      /* filter the data if necessary */
      %if (&&&type._numFilter ne 0 and 
           %bquote(&&&type._filter) ne %str()) %then 
      %do;      
         where = ((&&&type._rkVar in (%do f = 1 %to &&&type._numFilter;
                                         &&&type._filterRK&f
                                      %end;) and 
                   &&&type._filter)         
         %if (&&&type._numNoFilter ne 0) %then 
         %do;
             or (&&&type._rkVar in (%do f = 1 %to &&&type._numNoFilter;
                                         &&&type._noFilterRK&f
                                    %end;)) 
         %end;
         ) 
            
      %end; /* if - there is a filter */

   %mend pam_expand_transpose_filter;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_getData                               */
   /* Create the set statement to bring in the measure and/or event     */
   /* data with a by statement.  If both tables are available, then     */
   /* the data will be interleaved using the by statement.              */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_getData;

      %if (&measure and &event) %then
      %do;

         set &measure_inds (rename = (&measure_dttmVar = datetime)
                            /* apply where clause if necessary */
                            %pam_expand_transpose_filter(measure))
             &event_inds   (rename = (&event_dttmVar = datetime)
                            /* apply where clause if necessary */
                            %pam_expand_transpose_filter(event))
             end=eof;
         by &groupVar datetime;

      %end; /* if - both tables are available */

      %else %if (&measure) %then 
      %do;

         set &measure_inds (rename = (measure_dttm = datetime)
                            /* apply where clause if necessary */
                            %pam_expand_transpose_filter(measure))
             end=eof;
         by &groupVar datetime;

      %end; /* else if - just measure data */

      %else %if (&event) %then 
      %do;

         set &event_inds (rename = (event_dttm = datetime)
                          /* apply where clause if necessary */
                          %pam_expand_transpose_filter(event))
             end=eof;

         by &groupVar datetime;

      %end; /* else if - just event data */

   %mend pam_expand_transpose_getData;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_makeCol                               */
   /* Create the new columns in the transposed table                    */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_makeCol;

      /* create and retain each of the new transposed columns */
      retain %do m = 1 %to &measure_numUnique;
                &&measure_uniqueName&m
             %end;
             %do e = 1 %to &event_numUnique;
                &&event_uniqueNameN&e
                &&event_uniqueName&e
             %end;
      ;

      %if (&measure_numUnique gt 0) %then 
      %do;  
         format
         %do m = 1 %to &measure_numUnique;
            &&measure_uniqueName&m
         %end;
         nlnum8.4;
      %end;
      %if (&event_numUnique gt 0) %then 
      %do;  
         format
         %do e = 1 %to &event_numUnique;
            &&event_uniqueName&e
         %end;
         nlnum8.4
         %do e = 1 %to &event_numUnique;
            &&event_uniqueNameN&e
         %end;
         nlnum8.;
      %end;

   %mend pam_expand_transpose_makeCol;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_dtIf                                  */
   /* Create the if condition for the begin and end datetime range      */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_dtIf(and);
   
      &and "&enddttm"dt ge datetime ge "&begindttm"dt

   %mend pam_expand_transpose_dtIf;
   
   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_tsIf                                  */
   /* Create the if condition for the timeslice range.                  */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_tsIf(and);
   
      %if &timeslicebegin ne %str() %then 
      %do;
         &and "&timesliceend"t ge timepart(DATETIME) ge "&timeslicebegin"t
      %end; 

   %mend pam_expand_transpose_tsIf;
   
   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_groupIf                               */
   /* Create the if condition to check group variables                  */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_groupIf(m);

      %let and =;
      %do g = 1 %to &numGroup;

         %if (&g gt 1) %then 
            %let and = and;

         %if (&&groupVarType&g eq N) %then 
         %do;
            &and (&&groupVar&g eq &&group&g&m)
         %end;

         %else
         %do; 
            &and (&&groupVar&g eq "&&group&g&m")
         %end;

      %end; /* do g - loop through group variables */

   %mend pam_expand_transpose_groupIf;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_setMiss                               */
   /* Set the transposed columns to missing                             */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_setMiss(interpolate);
  
      drop miss;

      %if ((&interpolate eq ) or
           (&measure_int eq NONE)) %then 
      %do;
         call missing(miss
                      %do m = 1 %to &measure_numUnique;
                            , &&measure_uniqueName&m
                      %end; /* do i - loop through measures */
      %end; /* if - ignore interpolation - set everything to missing */
      %else
      %do;

         %if (&measure_int eq ) %then 
         %do; 
            /* if the interpoation method for a given group is NONE */
            /* then set the values to missing                       */
            %do m = 1 %to &measure_num;
               %if (&&measure_int&m eq NONE) %then 
               %do; 
                  if (%pam_expand_transpose_groupIf(&m)) then 
                     &&measure_name&m = .;
               %end;
            %end; /* do m - loop through measures */
         %end; /* if - override int is blank */

         /* start the call missing statement for events */
         call missing(miss

      %end;

      %do e = 1 %to &event_numUnique;
         , &&event_uniqueName&e 
      %end; /* do i - loop through events */
      );  /* end call missing statement */

      /* set event counts to 0 */
      %do e = 1 %to &event_numUnique;
         &&event_uniqueNameN&e = 0; 
      %end; /* do i - loop through events */

   %mend pam_expand_transpose_setMiss;   

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_accStat                               */
   /* Accumulate the values based on the specified statistic            */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_accStat(agg, stat, count, var);

      /* only accumulate non-missing values */
      if (&var ne .) then 
      do; 

         &count + 1;
         %if (&agg eq MEAN) %then 
         %do; 
            &stat=sum(&stat, &var);
         %end;
         %else %if (&agg eq SUM) %then 
         %do; 
            &stat=sum(&stat, &var);
         %end;
         %else %if (&agg eq MAX) %then 
         %do; 
            &stat=max(&stat, &var);
         %end;
         %else %if (&agg eq MIN) %then 
         %do; 
            &stat=min(&stat, &var);
         %end;
         %else %if (&agg eq LAST) %then 
         %do; 
            &stat=&var;
         %end;

      end; /* if - value is not missing */

   %mend pam_expand_transpose_accStat;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_finalStat                             */
   /* Final creation of the aggregated value                            */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_finalStat(agg, name, stat, count);

      /* need this check in case there are several rk's that have the */
      /* same transpose column name - don't want a missing value to   */
      /* overwrite the real value                                     */
      if (&name eq . and &stat ne .) then 
      do; 

         %if (&agg eq MEAN) %then 
         %do; 
            /* for MEAN - divide sum by count */
            if (&count gt 0 and &stat ne .) then 
               &name = &stat / &count;
            else
               &name = .;
         %end; /* if - agg is MEAN */

         %else %if (&agg eq NONE) %then 
         %do; 
            &name = .;
         %end;

         %else 
         %do; 
            &name = &stat;
         %end;

      end; /* if - name is missing and stat is not missing */

   %mend pam_expand_transpose_finalStat;

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_output                                */
   /* code to output records - either by expanding or aggregating       */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_output;

      /* flag to indicate the first datetime in the loop */
      /* - needed to help with interpolation             */
      first = 1;
      drop first;

      /* loop through and create records if gaps in the data */
      do datetime = old_dttm to enddttm 
                    by &periodicity;

         if (isAccumulating) then 
         do; 

            /* any final statistic calculations, and then set */
            /* transpose variables to the correct value       */
            %do m = 1 %to &measure_num;
               if (%pam_expand_transpose_groupIf(&m)) then 
               do;
                  %pam_expand_transpose_finalStat(&&measure_agg&m, 
                                                  &&measure_name&m, 
                                                  m_stat&m, m_count&m);
               end; /* if - correct group */
            %end;  /* do i - loop through measure columns */

            %do e = 1 %to &event_num;
               %pam_expand_transpose_finalStat(&&event_agg&e, 
                                               &&event_name&e, 
                                               e_stat&e, e_count&e);
            %end;  /* do i - loop through event columns */

         end; /* if - accumulating */

         else if (not first) then 
         do; 
            /* set transpose columns to missing */
            %pam_expand_transpose_setMiss;
         end; /* if - not the first record */

         first = 0;

         /* if there is a timeslice range then only output     */
         /* the record if it is in the range; otherwise output */
         if (1 %pam_expand_transpose_tsIf(and)) then 
         do; 

            /* if the aggregation is NONE then set values to missing */
            %do m = 1 %to &measure_num;
               %if (&&measure_agg&m eq NONE) %then 
               %do; 
                  if (%pam_expand_transpose_groupIf(&m)) then 
                     &&measure_name&m = .;
               %end;
            %end;  /* do m - loop through measure columns */
            %do e = 1 %to &event_num;
               %if (&&event_agg&e eq NONE) %then 
               %do; 
                  &&event_name&e = .;
               %end;
            %end;  /* do e - loop through event columns */
          
            %do m = 1 %to &measure_num;
               /* if interpolate is on and the transpose value is */
               /* missing, then set the transpose variable to the */
               /* interpolated value                              */
               %if (&&measure_int&m eq STEP) %then 
               %do; 
                  if (%pam_expand_transpose_groupIf(&m) and
                      &&measure_name&m eq .) then 
                  do; 
                     &&measure_name&m = m_int&m;
                  end;
               %end;
            %end;  /* do m - loop through measure columns */

            if (not firstRecord) then 
            do; 
               /* set DATETIME to the start of the interval */
               /* instead of the end; and then set it back  */
               /* to continue processing                    */
               datetime = datetime - &periodicity;
               output;
               datetime = datetime + &periodicity;

               /* count records in output table */
               numObs + 1;

            end;
            else
               firstRecord = 0;

            /* set interpolation values for next record */
            %do m = 1 %to &measure_num;
               m_int&m=&&measure_name&m;
            %end;

         end; /* if - record in timeslice range */

         /* once a record is output, accumulation stops */
         isAccumulating = 0;

      end; /* do datetime - loop through dates */

   %mend pam_expand_transpose_output; 

   /* ----------------------------------------------------------------- */
   /* Macro: pam_expand_transpose_drop                                  */
   /* Drop columns that should not be in the outds                      */
   /* ----------------------------------------------------------------- */
   %macro pam_expand_transpose_drop;

         drop &groupvar numObs;

         %if (&measure) %then 
         %do; 
            drop &measure_rkVar 
                 &measure_var;
         %end;
         %if (&event) %then 
         %do; 
            drop &event_rkVar 
                 &event_var;
         %end;

         /* drop the agg columns that don't have an aggregation   */
         /* method - only keep the count columns for these events */
         %do e = 1 %to &event_num;
            %if (&&event_agg&e eq NONE) %then 
            %do; 
                drop &&event_name&e;
            %end;
         %end; /* do e - loop through event columns */

   %mend pam_expand_transpose_drop;

   /* === P A R A M E T E R   V A L I D A T I O N === */

   %let primaryMacro = &sysmacroname;

   %pam_expand_transpose_nonBlank(beginDTTM);
   %pam_expand_transpose_nonBlank(endDTTM);

   %pam_expand_transpose_nonBlank(outds);

   /* periodicity must be a positive integer */
   %if (&periodicity eq ) %then 
      %let periodicity = 0;
   data _null_;
      periodicity = input("&periodicity", ? 10.);
      if (int(periodicity) ne periodicity or 
          periodicity eq . or 
          periodicity lt 0) then 
      do; 
         put "ERR" "OR: &primaryMacro requires a positive integer "
             "value for periodicity.";
         abort cancel 8;
      end;
   run;

   /* check that the measure_inds and the event_inds tables */
   /* exist and have data.  If not, set the _inds value to  */
   /* blank                                                 */
   %pam_expand_transpose_checkInds(measure);
   %pam_expand_transpose_checkInds(event);

   %if ((%nrquote(&measure_inds) eq ) and
        (%nrquote(&event_inds) eq )) %then 
   %do; 
      %let msg = &primaryMacro requires a value for either measure_inds; 
      %let msg = &msg or event_inds;
      %put ERROR%quote(:) &msg;
      %abort cancel 8;
   %end;

   /* flag to indicate if there is measure data */
   %let measure = 0;

   %if (%nrquote(&measure_inds) ne ) %then 
   %do;

      %let measure = 1;

      %pam_expand_transpose_dsExist(measure_inds);

      %pam_expand_transpose_setDefault(measure_dttmVar, measure_dttm);
      %pam_expand_transpose_setDefault(measure_rkVar, pam_tag_dim_rk);
      %pam_expand_transpose_setDefault(measure_var, loc_measure_value);

      %let dsid=%sysfunc(open(&measure_inds));
      %pam_expand_transpose_varExist(&dsid, measure_dttmVar);
      %pam_expand_transpose_varExist(&dsid, measure_rkVar);
      %pam_expand_transpose_varExist(&dsid, measure_var);
      %let dsid=%sysfunc(close(&dsid));

      %pam_expand_transpose_setDefault(measure_dim, &pam_dimLibref..pam_tag_dim);
      %pam_expand_transpose_dsExist(measure_dim);
      %pam_expand_transpose_setDefault(measure_colnameVar, tag_column_nm);
      %pam_expand_transpose_setDefault(measure_intVar, tag_def_interpol_method_cd);
      %pam_expand_transpose_setDefault(measure_aggVar, tag_def_aggreg_method_cd);

      /* make sure that measure_int has a valid value */
      %let measure_int = %kupcase(%nrquote(&measure_int));
      %if (%nrquote(&measure_int) eq DEFAULT) %then 
         %let measure_int = ;

      %if (%nrquote(&measure_int) ne STEP and
           %nrquote(&measure_int) ne NONE and
           %nrquote(&measure_int) ne ) %then 
      %do; 
         %let msg = &primaryMacro requires a valid value for;
         %let msg = &msg measure_int (DEFAULT, STEP, NONE, blank);
         %put ERROR%quote(:) &msg;
         %abort cancel 8;
      %end;

      %pam_expand_transpose_validAgg(measure_agg);

      %let dsid=%sysfunc(open(&measure_dim));
      %pam_expand_transpose_varExist(&dsid, measure_colnameVar);
      %pam_expand_transpose_varExist(&dsid, measure_intVar);
      %pam_expand_transpose_varExist(&dsid, measure_aggVar);
      %let dsid=%sysfunc(close(&dsid));

      %pam_expand_transpose_nonBlank(measure_rklist);
      %pam_expand_transpose_dsExist(measure_rklist);
      %pam_expand_transpose_setDefault(measure_idVar, tag_id);

      %let dsid=%sysfunc(open(&measure_rklist));
      %pam_expand_transpose_varExist(&dsid, measure_rkVar);
      %let dsid=%sysfunc(close(&dsid));

   %end; /* if - measure_inds is not blank */

   %let event = 0;
   %if (%nrquote(&event_inds) ne ) %then 
   %do;
      %let event = 1;

      %pam_expand_transpose_dsExist(event_inds);

      %pam_expand_transpose_setDefault(event_dttmVar, event_dttm);
      %pam_expand_transpose_setDefault(event_rkVar, pam_event_dim_rk);
      %pam_expand_transpose_setDefault(event_var, event_value);

      %let dsid=%sysfunc(open(&event_inds));
      %pam_expand_transpose_varExist(&dsid, event_dttmVar);
      %pam_expand_transpose_varExist(&dsid, event_rkVar);
      %pam_expand_transpose_varExist(&dsid, event_var);
      %let dsid=%sysfunc(close(&dsid));

      %pam_expand_transpose_validAgg(event_agg);

      %pam_expand_transpose_setDefault(event_dim, &pam_dimLibref..pam_event_dim);
      %pam_expand_transpose_dsExist(event_dim);
      %pam_expand_transpose_setDefault(event_colnameVar, event_column_nm);
      %pam_expand_transpose_setDefault(event_aggVar, event_def_aggreg_method_cd);

      %let dsid=%sysfunc(open(&event_dim));
      %pam_expand_transpose_varExist(&dsid, event_colnameVar);
      %pam_expand_transpose_varExist(&dsid, event_aggVar);
      %let dsid=%sysfunc(close(&dsid));

      %pam_expand_transpose_nonBlank(event_rklist);
      %pam_expand_transpose_dsExist(event_rklist);
      %pam_expand_transpose_setDefault(event_idVar, event_id);

      %let dsid=%sysfunc(open(&event_rklist));
      %pam_expand_transpose_varExist(&dsid, event_rkVar);
      %let dsid=%sysfunc(close(&dsid));

   %end; /* if - event_inds is not blank */

   %pam_expand_transpose_setDefault(groupVar, 
                                    %str(pam_asset_dim_rk pam_loc_dim_rk));

   /* make sure all group variables are in both input tables */
   /* and create groupVar macro variables                    */
   %if (&measure) %then 
      %let dsid_measure = %sysfunc(open(&measure_inds));
      
   %if (&event) %then 
      %let dsid_event = %sysfunc(open(&event_inds));
      
   %let lastGroupVar = ;
   %let count = 0;
   %global numGroup;
   %let numGroup = 0;
   
   %do %until(&gVar eq );
   
      %let count = %eval(&count + 1);
      %let gVar = %scan(%nrquote(&groupVar), &count, %str( ));
      
      %if (%nrquote(&gVar) ne ) %then 
      %do; 
      
         %let lastGroupVar = &gVar;
         %let msg = &primaryMacro requires that the variables for the;
         %let msg = &msg parameter groupVar exist in the;
         
         %if (&measure) %then 
         %do; 
         
            %let msg2 = &msg measure_inds (&measure_inds);
            %pam_expand_transpose_varExist(&dsid_measure, gVar, &msg2);
            %let numGroup = %eval(&numGroup + 1);
            %global groupVar&numGroup;
            %let groupVar&numGroup = &gvar;
            %let vnum = %sysfunc(varnum(&dsid_measure, &gVar));
            %global groupVarType&numGroup;
            %let groupVarType&numGroup = %sysfunc(vartype(&dsid_measure, &vnum));
            
         %end; /* if - there is measure data */
         
         %if (&event) %then 
         %do; 
            %let msg2 = &msg event_inds (&event_inds);
            %pam_expand_transpose_varExist(&dsid_event, gVar, &msg2);
         %end; /* if - there is event data */
                  
      %end; /* if - gVar is not blank */
      
   %end; /* do until - loop through group vars */ 

   %if (&measure) %then 
      %let dsid_measure=%sysfunc(close(&dsid_measure));
      
   %if (&event) %then 
      %let dsid_event=%sysfunc(close(&dsid_event));

   /* if there are filters, then create the filter and noFilter RK macro variables */
   %pam_expand_transpose_filterRK(measure);
   %pam_expand_transpose_filterRK(event);

   /* === P R O C E S S === */

   /* get all the information needed to do the transpose and */
   /* interpolation/aggregation                              */
   %let measure_num = 0;
   %let measure_numUnique = 0;
   %pam_expand_transpose_getColInfo(measure);
   
   %let event_num = 0;
   %let event_numUnique = 0;
   %pam_expand_transpose_getColInfo(event);

   %if (&periodicity eq 0) %then 
   %do; 

     /* --- Transpose only --- */

     data &outds;

         /* bring in measure and/or event data; interleave if both, and */
         /* apply filter(s) if necessary                                */
         %pam_expand_transpose_getData;

         /* create and retain each of the new transposed columns */
         %pam_expand_transpose_makeCol;

         if (first.datetime and not first.&lastGroupVar) then 
         do; 
            /* set transpose columns to missing unless interpolating */
            %pam_expand_transpose_setMiss(interpolate);
         end; /* if - first datetime */
         else if (first.&lastGroupVar) then 
         do; 
            /* set transpose columns to missing */
            %pam_expand_transpose_setMiss;
         end;

         %if (not &measure or &measure_int eq NONE) %then 
         %do; 
            /* if there is no interpolation, then keep only the  */
            /* datetimes in the begin-end range, and if there is */
            /* a timeslice range, keep only the data for it      */
            if (%pam_expand_transpose_dtIf
                %pam_expand_transpose_tsIf(and));
         %end;

         /* assign the value to its proper transpose variable */
         %let else=;
         %do m = 1 %to &measure_num;

            &else if (&measure_rkVar = &&measure_rk&m) then 
               &&measure_name&m = &measure_var;

            %let else = else;

         %end; /* do m - loop through measure columns */

         %do e = 1 %to &event_num;

            &else if (&event_rkVar = &&event_rk&e) then 
            do;
               &&event_name&e = &event_var;
               &&event_nameN&e = 1;
            end;

            %let else=else;

         %end; /* do e - loop through event columns */

         if (last.datetime
             %if (&measure_int ne NONE and &measure) %then 
             %do; 
                /* keep the data for the last datetime of the group, and  */
                /* only datetimes in the begin-end range, and if there is */
                /* a timeslice range, keep only the data for it           */
                and %pam_expand_transpose_dtIf
                    %pam_expand_transpose_tsIf(and)
             %end;
            ) then 
         do; 
            /* save record */
            output;

            /* count records in output table */
            numObs + 1;
         end;
 
         /* drop columns that shouldn't be in outds */
         %pam_expand_transpose_drop;

         /* save number of records */
         if (eof) then 
            call symputx("pam_fs_num_obs", numObs);
  
      run;

   %end; /* if - no periodicity */

   %else  
   %do; 

      /* --- Transpose and interpolate/aggregate --- */

      data &outds;

         /* bring in measure and/or event data; interleave if both, and */
         /* apply filter(s) if necessary                                */
         %pam_expand_transpose_getData;

         retain old_dttm isAccumulating 0;
         drop old_dttm isAccumulating;

         /* keep track if this record is before the begin-end datetime range */
         beforeRange = (datetime lt "&begindttm"dt);
         drop beforeRange;

         /* create and retain each of the new transposed columns */
         %pam_expand_transpose_makeCol;

         /* create and drop temporary columns to hold accumlated values */
         retain %do m = 1 %to &measure_num;
                   m_stat&m . m_count&m 0
                   m_int&m .
                %end; 
                %do e = 1 %to &event_num;
                   e_stat&e . e_count&e 0
                %end; 
         ;
         %if (&measure_num gt 0) %then 
         %do; 
            drop %do m = 1 %to &measure_num;
                    m_stat&m m_count&m
                    m_int&m
                 %end; 
                 ;
         %end;
         %if (&event_num gt 0) %then 
         %do; 
          
            drop %do e = 1 %to &event_num;
                    e_stat&e e_count&e
                 %end; 
                 ;
         %end;

         if (first.&lastGroupVar) then
         do; 
            /* set all transpose columns to missing */
            %pam_expand_transpose_setMiss;
            
            /* keep track that this is the first record of a group */ 
            retain firstRecord;
            drop firstRecord;
            firstRecord = 1;
            
            old_dttm = "&begindttm"dt;
            %do m = 1 %to &measure_num;
                m_int&m = .;
            %end; 

         end; /* if - first obs of group */

         /* if the datetime is greater than the last datetime, then   */
         /* need to output the last record, and start a new series of */
         /* records                                                   */
         if (datetime gt old_dttm) then 
         do; 

            enddttm = min(datetime-1, "&enddttm"dt);
            drop enddttm;

            %pam_expand_transpose_output;

            old_dttm = datetime;

            /* reset values */
            %pam_expand_transpose_setMiss;

         end; /* if - datetime is greater than previous datetime */

         /* if the datetime is <= the last datetime, then need to */
         /* accumulate the data (if data is not already being     */
         /* accumulated)                                          */
         if (datetime le old_dttm and not beforeRange and
             not isAccumulating) then 
         do;
 
            /* initialize the accumulation variables */
            %do m = 1 %to &measure_num;
               m_stat&m = .;
               m_count&m = 0;
            %end;
            %do e = 1 %to &event_num;
               e_stat&e = .;
               e_count&e = 0;
            %end;

            isAccumulating = 1;

         end; /* else if - datetime is less than previous datetime */

         /* assign the value to its proper accumulation or */
         /* interpolation variable                         */
         %let else=;
         %do m = 1 %to &measure_num;

            &else if (&measure_rkVar eq &&measure_rk&m) then 
            do; 
               if (isAccumulating) then 
               do; 
               
                  /* check that record is the right group values for */
                  /* this aggregation method                         */                    
                  if (%pam_expand_transpose_groupIf(&m)) then 
                  do;
                     /* accumulate the stats appropriately */
                     %pam_expand_transpose_accStat(&&measure_agg&m, m_stat&m, 
                                                   m_count&m, &measure_var);
                  end;  
                     
               end; /* if - accumulating */
               else
                  m_int&m = &measure_var; 

            end; /* else if - this measure */
            %let else=else;

         %end; /* do m - loop through measure columns */

         %do e = 1 %to &event_num;

            &else if (&event_rkVar eq &&event_rk&e) then 
            do; 
               if (isAccumulating) then 
               do; 
                  /* accumulate the stats appropriately */
                  %pam_expand_transpose_accStat(&&event_agg&e, e_stat&e, 
                                                e_count&e, &event_var);
                  &&event_nameN&e = &&event_nameN&e + 1;
               end; /* if - currently accumulating */
               else if (not beforeRange) then 
               do;
                  &&event_name&e=&event_var;
                  &&event_nameN&e = 1;
               end;

            end; /* else if - this event */
            %let else=else;

         %end; /* do e - loop through event columns */

         /* output the last columns in this group */
         if (last.&lastGroupVar) then 
         do; 
            enddttm = "&enddttm"dt;
            drop enddttm;
            %pam_expand_transpose_output;
         end;

         /* drop columns that should not be in outds */
         %pam_expand_transpose_drop;

         /* save number of records */
         if (eof) then 
            call symputx("pam_fs_num_obs", numObs);

      run;

   %end; /* else - periodicity is at least 1 */

   %if (&debug ne 1) %then 
   %do; 

      /* delete global macro variables */
      %do m = 1 %to &measure_num;
         %symdel measure_id&m measure_rk&m measure_name&m 
                 measure_int&m measure_agg&m 
                 / nowarn;
         %do g = 1 %to &numGroup;
            %symdel group&g&m / nowarn;
         %end;
      %end;

      %do e = 1 %to &event_num;
         %symdel event_id&e event_rk&e event_name&e 
                 event_nameN&e event_agg&e 
                 / nowarn;
      %end;

      %symdel measure measure_num event event_num / nowarn;

      %do m = 1 %to &measure_numUnique;
         %symdel measure_UniqueName&m /nowarn;
      %end;
      
      %do e = 1 %to &event_numUnique;
         %symdel event_UniqueName&e event_UniqueNameN&e / nowarn;
      %end;
      
      %symdel measure_numUnique event_numUnique / nowarn;

      %do g = 1 %to &numGroup;
         %symdel groupVar&g groupVarType&g / nowarn;
      %end;

      %symdel numGroup / nowarn;
      
   %end /* if - debug is not on */;

   %put %str( );
   %put *** &sysmacroname complete ***;
   %put %str( );

%mend pam_expand_transpose;
