%macro pam_gen_fs_get_data_sel(data_selection=);
/* Get data selection parts from data_sel domain object table. */
data _null_;
  set &pam_dolibref..data_sel;
  if data_sel_id eq "&data_selection";
  
  call symput('periodicityInterval',periodicity_interval_cd);
  call symput('periodicity',put(interpolation_granularity_no,8.));
  call symput('filteredSubsetTableName',data_sel_id);
  
  if data_sel_transpose_flg ne "" then do;
    call symput('transposeFlag',data_sel_transpose_flg);
  end;
  else do;
    call symput('transposeFlag',"BLANK");
  end;

  if kupcase(data_sel_interpolate_flg) eq "Y" then do;
    call symput('interpolationMethodTag',kupcase(interpolation_method_cd));
  end;
  else do;
    call symput('interpolationMethodTag',"");
  end;

  if kupcase(data_sel_aggregate_flg) eq "Y" then do;
    call symput('aggregationMethodTag',kupcase(aggregation_method_tag_cd));
    call symput('aggregationMethodEvent',kupcase(aggregation_method_event_cd));

    if periodicity_interval_cd ne "" and interpolation_granularity_no ne . then do;
      if kupcase(periodicity_interval_cd) eq "SECONDS" then do;
	    periodicitySeconds = interpolation_granularity_no;
	  end;
      else if kupcase(periodicity_interval_cd) eq "HOURS" then do;
        periodicitySeconds = interpolation_granularity_no * 3600;
	  end;
      else if kupcase(periodicity_interval_cd) eq "DAYS" then do;
        periodicitySeconds = interpolation_granularity_no * 86400;
	  end;
      call symput('periodicitySeconds', put(periodicitySeconds,8.));
    end;
  end;
  else do;
    call symput('aggregationMethodTag',"");
    call symput('aggregationMethodEvent',"");
    call symput('periodicitySeconds',"");
  end;

  if period_interval_cd ne "" and period_cnt ne 0 then do;
    call symput('dynamic',"Y");
    call symput('p_period_interval_cd',period_interval_cd);
    call symput('p_period_count',put(period_cnt,8.));
  end;
  else do;
    call symput('dynamic',"N");
    call symput('startDatetime',put(data_sel_start_dttm,datetime.));
    call symput('endDatetime',put(data_sel_end_dttm,datetime.));
    call symput('startDate',put(datepart(data_sel_start_dttm),date9.));
    call symput('endDate',put(datepart(data_sel_end_dttm),date9.));
  end;

  if data_sel_start_tm ne . and data_sel_end_tm ne . then do;
    call symput('startTimeSlice',put(data_sel_start_tm,time8.));
    call symput('endTimeSlice',put(data_sel_end_tm,time8.));
  end;
  else do;
    call symput('startTimeSlice',put('00:00:00't,time8.));
    call symput('endTimeSlice',put('23:59:59't,time8.));
  end;
run;
%if &dynamic eq Y %then %do;
  %pam_resolve_dynamic_dt(data_sel=&data_selection);
%end;
%mend;

%macro pam_gen_fs_get_dim_table_names;
/* Get the table names for the Asset and Location dimension tables. */
%local locTableId;
proc sql noprint;
  select table_id into :locTableId
  from &pam_mfglibref..table_attributes
  where kupcase(attribute_nm) eq "DIMENSION_TYPE" and
        kupcase(attribute_value) eq "LOCATION";
quit;
proc sql noprint;
  select table_nm into :locDimTable
  from &pam_mfglibref..table_meta
  where table_id eq &locTableId;
quit;

%local assetTableId;
proc sql noprint;
  select table_id into :assetTableId
  from &pam_mfglibref..table_attributes
  where kupcase(attribute_nm) eq "DIMENSION_TYPE" and
        kupcase(attribute_value) eq "ASSET";
quit;
proc sql noprint;
  select table_nm into :assetDimTable
  from &pam_mfglibref..table_meta
  where table_id eq &assetTableId;
quit;
%mend;

%macro pam_gen_fs_get_content(contentTypes=, dataSelection=);
%if %nrquote(&contentTypes) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_content requires a value for parameter contentTypes;
  %let pam_gen_fs_rc = 8;
%end;
%if %nrquote(&dataSelection) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_content requires a value for parameter dataSelection;
  %let pam_gen_fs_rc = 8;
%end;

/* Create a macro variable for each content type. */
%local numContentTypes;
%let numContentTypes = %eval(%sysfunc(countc(%nrbquote(&contentTypes), %str( )))+1);

%do i=1 %to &numContentTypes&;
  %let contentType&i=%sysfunc(scan(&contentTypes, &i));
  %if &&contentType&i eq EVENTS %then %let longContentType&i = MAINEVENTSOURCE; 
  %if &&contentType&i eq TAGS %then %let longContentType&i = MAINSENSORSOURCE; 
%end;                                                    
 
/* Create a data set work.data_sel_<contentType> for each Content Type. */
%do i=1 %to &numContentTypes;
  proc sql noprint;
    create table work.data_sel_&&contentType&i as
      select * 
      from &pam_dolibref..data_sel_content
      where kupcase(content_type_cd) eq kupcase("&&longContentType&i") and
          data_sel_id eq "&dataSelection";
  quit;
%end;

%mend;

%macro pam_gen_fs_get_filter_content(filterContentTypes=, dataSelection=);
%if %nrquote(&filterContentTypes) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs requires a value for parameter filterContentTypes;
  %let pam_gen_fs_rc = 8;
%end;
%if %nrquote(&dataSelection) eq   %then %do;
  %put ERROR%quote(:) pam_gen_fs_get_filter_content requires a value for parameter dataSelection;
  %let pam_gen_fs_rc = 8;
%end;

/* Create a macro variable for each content type. */
%local numFilterContentTypes;
%let numFilterContentTypes = %eval(%sysfunc(countc(%nrbquote(&filterContentTypes), %str( )))+1);

%do i=1 %to &numFilterContentTypes&;
  %let filterContentType&i=%sysfunc(scan(&filterContentTypes, &i));
  %if &&filterContentType&i eq EVENTS %then %let longFilterContentType&i = MAINEVENTSOURCE; 
  %if &&filterContentType&i eq TAGS %then %let longFilterContentType&i = MAINSENSORSOURCE;  
%end;                                                    
 
/* Create a data set work.data_sel_<filterContentType>_data_filter for each Filter Content Type. */
%do i=1 %to &numFilterContentTypes;

  proc sql noprint;
    create table work.data_sel_&&filterContentType&i.._data_filter as
      select * 
      from &pam_dolibref..data_sel_content_data_filter
      where kupcase(content_type_cd) eq kupcase("&&longFilterContentType&i") and
          data_sel_id eq "&dataSelection";
  quit;
%end;

%mend;

%macro pam_gen_fs_get_mart_meta_tag;
  proc sql noprint;
    create table tableAttr1 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "TABLE_TYPE" and 
          kupcase(attribute_value) eq "CONTENT" ;
  quit;
  proc sql noprint;
    create table tableAttr2 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "CONTENT_TYPE" and 
          kupcase(attribute_value) eq "MAINSENSORSOURCE";
  quit;
  proc sql noprint;
    select distinct(tableMeta.table_nm) into :tagDimTable
    from &pam_mfglibref..table_meta as tableMeta, 
         work.tableAttr1 as attr1, 
         work.tableAttr2 as attr2
    where attr1.table_id eq attr2.table_id and 
          attr1.table_id eq tableMeta.table_id;
  quit;

  proc sql noprint; 
    create table work.tempTableAttr1 as
    select *  
    from &pam_mfglibref..table_attributes as table_attributes 
    where kupcase(table_attributes.attribute_nm) eq "TABLE_TYPE" and
          kupcase(table_attributes.attribute_value) eq "SUBJECT";
  quit;
  proc sql noprint; 
    create table work.tempTableAttr2 as
    select *  
    from &pam_mfglibref..table_attributes as table_attributes 
    where kupcase(table_attributes.attribute_nm) eq "SUBJECT_TYPE" and
          kupcase(table_attributes.attribute_value) eq "TAGS";
  quit;
  proc sql noprint;
    select table_meta.table_nm into :measureFactTable
    from work.tempTableAttr1 as table_attr1,
         work.tempTableAttr2 as table_attr2,
         &pam_mfglibref..table_meta as table_meta
    where table_attr1.table_id eq table_attr2.table_id and
          table_attr1.table_id eq table_meta.table_id;
  quit;

  %if &p_table_names ne  %then %do;
    %let p_table_names = &p_table_names. &measureFactTable;
  %end;
  %else %do;
    %let p_table_names = &measureFactTable;
  %end;

  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: measure_val_col_nm
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes,
         &pam_mfglibref..table_attributes as table_attributes
    where tablecolumn_meta.table_id eq tablecolumn_attributes.table_id and
          tablecolumn_attributes.table_id eq table_attributes.table_id and
          tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          kupcase(table_attributes.attribute_value) eq "TAGS" and
          kupcase(tablecolumn_attributes.attribute_value) eq "MEASURE_VALUE";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: measure_dttm_col_nm
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes,
         &pam_mfglibref..table_attributes as table_attributes
    where tablecolumn_meta.table_id eq tablecolumn_attributes.table_id and
          tablecolumn_attributes.table_id eq table_attributes.table_id and
          tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          kupcase(table_attributes.attribute_value) eq "TAGS" and
          kupcase(tablecolumn_attributes.attribute_value) eq "MEASURE_DTTM";
  quit;
  proc sql noprint;
    select table_id 
    into :tag_table_id
    from &pam_mfglibref..table_attributes as table_attributes
    where kupcase(table_attributes.attribute_nm) eq "CONTENT_TYPE" and
          kupcase(table_attributes.attribute_value) eq "MAINSENSORSOURCE";
  quit;
  
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: RKColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &tag_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "RETAINED_KEY";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: tagIdColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &tag_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "BUSINESS_KEY";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: tagTransposeColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &tag_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "TRANSPOSE_COLUMN";
  quit;
%mend;

%macro pam_gen_fs_get_mart_meta_event;
  proc sql noprint;
    create table tableAttr1 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "TABLE_TYPE" and 
          kupcase(attribute_value) eq "CONTENT" ;
  quit;
  proc sql noprint;
    create table tableAttr2 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "CONTENT_TYPE" and 
          kupcase(attribute_value) eq "MAINEVENTSOURCE";
  quit;
  proc sql noprint;
    select distinct(tableMeta.table_nm) into :eventDimTable
    from &pam_mfglibref..table_meta as tableMeta, 
         work.tableAttr1 as attr1, 
         work.tableAttr2 as attr2
    where attr1.table_id eq attr2.table_id and 
          attr1.table_id eq tableMeta.table_id;
  quit;

  proc sql noprint;
    create table tableAttr1 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "TABLE_TYPE" and 
          kupcase(attribute_value) eq "CONTENT" ;
  quit;
  proc sql noprint;
    create table tableAttr2 as
    select table_id from &pam_mfglibref..table_attributes
    where kupcase(attribute_nm) eq "CONTENT_TYPE" and 
          kupcase(attribute_value) eq "MAINASSETSOURCE";
  quit;
  proc sql noprint;
    select distinct(tableMeta.table_nm) into :assetDimTable
    from &pam_mfglibref..table_meta as tableMeta, 
         work.tableAttr1 as attr1, 
         work.tableAttr2 as attr2
    where attr1.table_id eq attr2.table_id and 
          attr1.table_id eq tableMeta.table_id;
  quit;

  proc sql noprint; 
    create table work.tempTableAttr1 as
    select *  
    from &pam_mfglibref..table_attributes as table_attributes 
    where kupcase(table_attributes.attribute_nm) eq "TABLE_TYPE" and
          kupcase(table_attributes.attribute_value) eq "SUBJECT";
  quit;
  proc sql noprint; 
    create table work.tempTableAttr2 as
    select *  
    from &pam_mfglibref..table_attributes as table_attributes 
    where kupcase(table_attributes.attribute_nm) eq "SUBJECT_TYPE" and
          kupcase(table_attributes.attribute_value) eq "EVENTS";
  quit;
  proc sql noprint;
    select table_meta.table_nm into :eventFactTable
    from work.tempTableAttr1 as table_attr1,
         work.tempTableAttr2 as table_attr2,
         &pam_mfglibref..table_meta as table_meta
    where table_attr1.table_id eq table_attr2.table_id and
          table_attr1.table_id eq table_meta.table_id;
  quit;

  %if &p_table_names ne  %then %do;
    %let p_table_names = &p_table_names. &eventFactTable;
  %end;
  %else %do;
    %let p_table_names = &eventFactTable;
  %end;

  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: event_value_col_nm
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes,
         &pam_mfglibref..table_attributes as table_attributes
    where tablecolumn_meta.table_id eq tablecolumn_attributes.table_id and
          tablecolumn_attributes.table_id eq table_attributes.table_id and
          tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          kupcase(table_attributes.attribute_value) eq "EVENTS" and
          kupcase(tablecolumn_attributes.attribute_value) eq "MEASURE_VALUE";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: event_dttm_col_nm
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes,
         &pam_mfglibref..table_attributes as table_attributes
    where tablecolumn_meta.table_id eq tablecolumn_attributes.table_id and
          tablecolumn_attributes.table_id eq table_attributes.table_id and
          tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          kupcase(table_attributes.attribute_value) eq "EVENTS" and
          kupcase(tablecolumn_attributes.attribute_value) eq "MEASURE_DTTM";
  quit;
  proc sql noprint;
    select table_id 
    into :event_table_id
    from &pam_mfglibref..table_attributes as table_attributes
    where kupcase(table_attributes.attribute_nm) eq "CONTENT_TYPE" and
          kupcase(table_attributes.attribute_value) eq "MAINEVENTSOURCE";
  quit;
  
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: eventRKColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &event_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "RETAINED_KEY";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: eventIdColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &event_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "BUSINESS_KEY";
  quit;
  proc sql noprint;
    select distinct tablecolumn_meta.column_nm
    into: eventTransposeColumnName
    from &pam_mfglibref..tablecolumn_meta as tablecolumn_meta,
         &pam_mfglibref..tablecolumn_attributes as tablecolumn_attributes
    where tablecolumn_attributes.column_id eq tablecolumn_meta.column_id and
          tablecolumn_meta.table_id eq &event_table_id and
          kupcase(tablecolumn_attributes.attribute_value) eq "TRANSPOSE_COLUMN";
  quit;
%mend;

%macro pam_gen_fs_get_list_of_tags(data_selection=);
proc sql noprint;
  create table work.data_sel_tags as
  select distinct content_id as tag_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINSENSORSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;
proc sql noprint;
  create table work.tagIDs as
  select data_sel_fact.fact_tbl_cd, data_sel_tags.tag_id
  from work.data_sel_tags as data_sel_tags,
       &pam_dolibref..data_sel as data_sel,
       &pam_dolibref..data_sel_fact as data_sel_fact
  where data_sel_fact.data_sel_id eq data_sel.data_sel_id and
        data_sel.data_sel_id eq "&data_selection" and
        kupcase(data_sel_fact.fact_tbl_cd) eq "TAGS";
quit;

proc sql noprint;
  select count(distinct tag_id) into :ntags
  from work.tagIDs;
quit;

%let ntags=%cmpres(&ntags);

%do i=1 %to &ntags;
  %global tagList&i;
%end;

proc sql noprint;
  select distinct quote(ktrim(kleft(tag_id)))
  into :tagList1 - :tagList&ntags
  from work.tagIDs;
quit;

%global tagListLength;
%let tagListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  create table tagIDsAndRKs as
  select distinct tagIDS.tag_id, &RKColumnName
  from work.tagIDs as tagIDS, 
       &pam_dimlibref..&tagDimTable as tagDim
  where tagIDs.tag_id eq tagDim.tag_id and current_num eq 1;
quit;
data tagIDToRKFmt(keep=fmtname start label type);
  set work.tagIDsAndRKs;
  retain fmtname "tagIDToRKfmt" type "C";
  start=tag_id;
  label=&RKColumnName;
run;
proc sort data=work.tagIDToRKFmt nodupkey;
  by start label;
run;
proc format cntlin=tagIDToRKFmt;
run;
%mend;

/* Gets the RKs for all tags in the list of selected tags. */
%macro pam_gen_fs_get_tag_RKs;
proc sql noprint;
  select count(distinct &RKColumnName) into :ntagRKs
  from &pam_dimlibref..&tagDimTable
  where &tagIdColumnName in (
     %do i=1 %to %eval(&tagListLength. -1);
        &&tagList&i,
     %end;
        &&tagList&tagListLength) 
     and current_num eq 1;
quit;

%let ntagRKs=%cmpres(&ntagRKs);

%do i=1 %to &ntagRKs;
  %global tagRKList&i;
%end;

proc sql noprint;
  select &RKColumnName into :tagRKList1 - :tagRKList&ntagRKs
  from &pam_dimlibref..&tagDimTable
  where &tagIdColumnName in (
     %do i=1 %to %eval(&tagListLength. -1);
        &&tagList&i,
     %end;
        &&tagList&tagListLength)
     and current_num eq 1;
quit;
%global tagRKListLength;
%let tagRKListLength = %sysfunc(compress(&sqlobs));
%mend;

/* Perform the data selection against the fact table. */
%macro pam_gen_fs_get_extra_tag_obs;
data _null_;
  set &pamapp_libref..fact_tag_measures_stats(where=(&RKColumnName in (
    %do i=1 %to %eval(&tagRKListLength. -1);
      &&tagRKList&i,
    %end;
      &&tagRKList&tagRKListLength )));

  by &RKColumnName tag_measure_dt;

  retain startDate endDate laterDate1 laterDate2 prevDate1 prevDate2 foundStart foundEnd 0;
  format startDate endDate laterDate1 laterDate2 prevDate1 prevDate2 date9.;

  if _n_ eq 1 then do;
    startDate = input("&startDate",date9.);
    endDate = input("&endDate",date9.);
  end;

  startDateTemp = input("&startDate",date9.);
  endDateTemp = input("&endDate",date9.);

  if tag_measure_dt eq startDateTemp then foundStart = 1;
  if tag_measure_dt eq endDateTemp then foundEnd = 1;

  if tag_measure_dt < startDateTemp then do;
    if prevDate1 eq . then prevDate1 = tag_measure_dt;
    if prevDate2 eq . then prevDate2 = tag_measure_dt;
    if tag_measure_dt > prevDate1 then do;
      prevDate2 = prevDate1;
      prevDate1 = tag_measure_dt;
    end;
  end;
  if tag_measure_dt > endDateTemp then do;
    if laterDate1 eq . then laterDate1 = tag_measure_dt;
    if laterDate2 eq . then laterDate2 = tag_measure_dt;
    if tag_measure_dt > laterDate2 then do;
      laterDate1 = laterDate2;
      laterDate2 = tag_measure_dt;
    end;
  end;

  if last.&RKColumnName then do;
    if not foundStart then do;
      if prevDate2 < startDateTemp then startDateTemp = prevDate2;
    end;
    else do;
      if prevDate1 < startDateTemp then startDateTemp = prevDate1;
    end;
    if not foundEnd then do;
      if laterDate2 > endDateTemp then endDateTemp = laterDate2;
    end;
    else do;
      if laterDate1 > endDateTemp then endDateTemp = laterDate1;
    end;
    foundStart = 0;
    foundEnd = 0;
    prevDate1 = .;
    prevDate2 = .;
    laterDate1 = .;
    laterDate2 = .;
  end;

  if startDateTemp < startDate then startDate = startDateTemp;
  if endDateTemp > endDate then endDate = endDateTemp;

  call symput('startDate',put(startDate,date9.));
  call symput('endDate',put(endDate,date9.));
run;
%mend;

%macro pam_gen_fs_get_measure_facts;
  proc sql noprint;
    create table work.AssetAndTagRKs as
    select &RKColumnName, pam_asset_dim_rk
    from &pam_dimlibref..&tagDimTable
    where &RKColumnName in (
    %do i=1 %to %eval(&tagListLength. -1);
      &&tagRKList&i,
    %end;
      &&tagRKList&tagRKListLength);
  quit;
  proc sql noprint;
    create table work.AssetIdsAndTagRKs as
    select &RKColumnName, pam_asset_id
    from work.AssetAndTagRKs as RKs,
         &pam_dimlibref..&assetDimTable as AssetDim
    where RKs.pam_asset_dim_rk eq AssetDim.pam_asset_dim_rk;
  quit;
  data tagToAssetIdFmt(keep=fmtname start label type);
    set work.AssetIdsAndTagRKs;
    retain fmtname "tagToAssetIdFmt" type "N";
    start=&RKColumnName;
    label=pam_asset_id;
  run;
  proc sort data=work.tagToAssetIdFmt;
    by start label;
  run;
  proc format cntlin=tagToAssetIdFmt;
  run;
  data tagToAssetRKFmt(keep=fmtname start label type);
    set work.AssetAndTagRKs;
    retain fmtname "tagToAssetRKFmt" type "N";
    start=&RKColumnName;
    label=pam_asset_dim_rk;
  run;
  proc sort data=work.tagToAssetRKFmt;
    by start label;
  run;
  proc format cntlin=tagToAssetRKFmt;
  run;
  proc sql noprint;
    create table work.tagData as
    select put(pam_loc_dim_rk,locfmt.) as LOC_CD length=20, 
           put(&RKColumnName,tagToAssetIdFmt.) as ASSET_ID length=32,
           input(put(&RKColumnName,tagToAssetRKFmt.),best8.) as PAM_ASSET_DIM_RK, 
           &RKColumnName, 
           pam_loc_dim_rk, 
           &measure_dttm_col_nm, 
           &measure_val_col_nm
    from &pam_meas_fact_libref..&measureFactTable
    where &RKColumnName in (
    %do i=1 %to %eval(&tagRKListLength. -1);
      &&tagRKList&i,
    %end;
      &&tagRKList&tagRKListLength) 
      and pam_date_dim_rk between "&startdate"d and "&enddate"d
    order by pam_asset_dim_rk, loc_cd, &measure_dttm_col_nm;
  quit;
  %if not(&sqlobs) %then %do;
    proc sql noprint;
      create table work.tagData (
        pam_loc_dim_rk numeric,
        &RKColumnName numeric,
        pam_asset_dim_rk numeric,
        &measure_dttm_col_nm numeric,
        &measure_val_col_nm numeric,
        LOC_CD varchar(20),
        ASSET_ID varchar(32)
	  );
    quit;
    %put;
    %put WARNING:  The current data selection definition results in no selected ;
    %put WARNING:  tag data.  You must change your data selection definition or ;
    %put WARNING:  load data into your data mart and re-run the data selection ;
    %put WARNING:  to see tag data in the filtered subset. ;
    %put;
    %let pam_gen_fs_rc = 4;
  %end;
%mend;

%macro pam_gen_fs_get_list_of_assets(data_selection=);
proc sql noprint;
  create table work.data_sel_assets as
  select distinct content_id as asset_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINASSETSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;
%if not(&sqlobs) %then %do;
  proc sql noprint;
    select count(distinct pam_asset_id) into :nassets
    from &pam_dimlibref..&assetDimTable;
  quit;

  %let nassets=%cmpres(&nassets);

  %do i=1 %to &nassets;
    %global assetList&i;
  %end;

  proc sql noprint;
    select distinct quote(ktrim(kleft(asset_id)))
    into :assetList1 - :assetList&nassets
    from &pam_dimlibref..&assetDimTable;
  quit;

  %global assetListLength;
  %let assetListLength = %sysfunc(compress(&sqlobs));

  %put;
  %put WARNING:  The current data selection definition contains no selected ;
  %put WARNING:  asset IDs.  All asset IDs in the asset dimension table in  ;
  %put WARNING:  the data mart will be used. ;
  %put;
  %let pam_gen_fs_rc = 4;
%end;
%else %do;
  proc sql noprint;
    select count(distinct asset_id) into :nassets
    from work.data_sel_assets;
  quit;

  %let nassets=%cmpres(&nassets);

  %do i=1 %to &nassets;
    %global assetList&i;
  %end;
 
  proc sql noprint;
    select distinct quote(ktrim(kleft(asset_id)))
    into :assetList1 - :assetList&nassets
    from work.data_sel_assets;
  quit;
 
  %global assetListLength;
  %let assetListLength = %sysfunc(compress(&sqlobs));
%end;
%mend;

%macro pam_gen_fs_get_list_of_events(data_selection=);

proc sql noprint;
  create table work.data_sel_events as
  select distinct content_id as event_id
  from &pam_dolibref..data_sel_content
  where kupcase(content_type_cd) eq "MAINEVENTSOURCE" and 
        kupcase(data_sel_id) eq kupcase("&data_selection");
quit;

proc sql noprint;
  create table work.eventIDs as
  select data_sel_fact.fact_tbl_cd, data_sel_events.event_id
  from work.data_sel_events as data_sel_events,
       &pam_dolibref..data_sel as data_sel,
       &pam_dolibref..data_sel_fact as data_sel_fact
  where data_sel_fact.data_sel_id eq data_sel.data_sel_id and
        data_sel.data_sel_id eq "&data_selection" and
        kupcase(data_sel_fact.fact_tbl_cd) eq "EVENTS";
quit;

proc sql noprint;
  select count(distinct event_ID) into :nevents
  from work.eventIDs;
quit;

%let nevents=%cmpres(&nevents);

%do i=1 %to &nevents;
  %global eventList&i;
%end;

proc sql noprint;
  select distinct quote(ktrim(kleft(event_id)))
  into :eventList1 - :eventList&nevents 
  from work.eventIDs;
quit;

%global eventListLength;
%let eventListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  select count(distinct &eventRKColumnName) into :neventRKs
  from &pam_dimlibref..&eventDimTable
  where &eventIdColumnName in (
     %do i=1 %to %eval(&eventListLength. -1);
        &&eventList&i,
     %end;
        &&eventList&eventListLength) 
     and current_num eq 1;
quit;

%let neventRKs=%cmpres(&neventRKs);

%do i=1 %to &neventRKs;
  %global eventRKList&i;
%end;

proc sql noprint;
  select &eventRKColumnName into :eventRKList1 - :eventRKList&neventRKs 
  from &pam_dimlibref..&eventDimTable
  where &eventIdColumnName in (
     %do i=1 %to %eval(&eventListLength. -1);
        &&eventList&i,
     %end;
        &&eventList&eventListLength) 
     and current_num eq 1;
quit;
%global eventRKListLength;
%let eventRKListLength = %sysfunc(compress(&sqlobs));

proc sql noprint;
  create table work.eventColumnNames as
  select eventIDs.event_id, pam_event_dim.&eventTransposeColumnName, pam_event_dim.&eventRKColumnName
  from work.eventIDs as eventIDs,
       &pam_dimlibref..&eventDimTable as pam_event_dim
  where eventIDs.event_id eq pam_event_dim.event_id and current_num eq 1;
quit;

data eventIDToRKFmt(keep=fmtname start label type);
  set work.eventColumnNames;
  retain fmtname "eventIDToRKfmt" type "C";
  start=event_id;
  label=&eventRKColumnName;
run;
proc sort data=work.eventIDToRKFmt;
  by start label;
run;
proc format cntlin=eventIDToRKFmt;
run;

%mend;

/* Gets the RKs for all assets in the list of selected assets. */
%macro pam_gen_fs_get_asset_RKs;
proc sql noprint;
  select count(distinct pam_asset_dim_rk) into :nassetRKs
  from &pam_dimlibref..&assetDimTable
  where current_num eq 1;
quit;

%let nassetRKs=%cmpres(&nassetRKs);

%do i=1 %to &nassetRKs;
  %global assetRKList&i;
%end;

proc sql noprint;
  select distinct pam_asset_dim_rk
  into :assetRKList1 - :assetRKList&nassetRKs
  from &pam_dimlibref..&assetDimTable
  where
    %if &assetList1 ne %str() %then %do;
      pam_asset_id in (
        %do i=1 %to %eval(&assetListLength. -1);
          &&assetList&i,
        %end;
          &&assetList&assetListLength) and
    %end;
    current_num eq 1;
quit;

%global assetRKListLength;
%let assetRKListLength = %sysfunc(compress(&sqlobs));

%mend;

%macro pam_gen_fs_get_event_facts;
  proc sql noprint;
    create table work.assetIDs as
    select distinct pam_asset_dim_rk, pam_asset_ID
    from &pam_dimlibref..&assetDimTable
    where pam_asset_dim_rk in (
       %do i=1 %to %eval(&assetRKListLength. -1);
         &&assetRKList&i,
       %end;
         &&assetRKList&assetRKListLength)
       and current_num=1;
  quit;
  data assetRKfmt(keep=start label fmtname type);
    set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label));
    retain fmtname "assetRKfmt" type "N";
  run;
  proc sort data=work.assetRKfmt;
    by start label;
  run;
  proc format cntlin=work.assetRKfmt;
  run;

  proc sql noprint;
    create table work.eventData as
    select put(pam_loc_dim_rk,locfmt.) as LOC_CD length=20, 
           put(pam_asset_dim_rk,assetRKfmt.) as ASSET_ID length=32,
           &eventRKColumnName, 
           pam_asset_dim_rk, 
           pam_loc_dim_rk, 
           &event_dttm_col_nm, 
           &event_value_col_nm
    from &pam_meas_fact_libref..&eventFactTable
    where &eventRKColumnName in (
       %do i=1 %to %eval(&eventRKListLength. -1);
         &&eventRKList&i,
       %end;
         &&eventRKList&eventRKListLength) and
      %if &assetRKList1 ne %str() %then %do;
          pam_asset_dim_rk in (
             %do i=1 %to %eval(&assetRKListLength. -1);
                &&assetRKList&i,
             %end;
                &&assetRKList&assetRKListLength) and
      %end;
          pam_date_dim_rk between "&startdate"d and "&enddate"d
    order by pam_asset_dim_rk, loc_cd, &event_dttm_col_nm;
  quit;
  %if not(&sqlobs) %then %do;
    proc sql noprint;
      create table work.eventData (
        pam_loc_dim_rk numeric,
        &eventRKColumnName numeric,
        pam_asset_dim_rk numeric,
        &event_dttm_col_nm numeric,
        &event_value_col_nm numeric,
        LOC_CD varchar(20),
        ASSET_ID varchar(32)
      );
    quit;

    %put;
    %put WARNING:  The current data selection definition results in no selected ;
    %put WARNING:  event data.  You must change your data selection definition or ;
    %put WARNING:  load data into your data mart and re-run the data selection ;
    %put WARNING:  to see event data in the filtered subset. ;
    %put;
    %let pam_gen_fs_rc = 4;
  %end;
%mend;

%macro pam_gen_fs_transpose(data_selection=);
%local tagRKsNoFilter1;
%local eventRKsNoFilter1;
%local tagFilter;
%local eventFilter;

%if &extractMeasures eq Y %then %do;
  proc sql noprint;
    create table work.data_sel_tags_data_filter as
    select data_sel_id, content_type_cd, content_id as tag_id, filter_order_no, filter_value_no, 
           filter_operator_cd, filter_logical_operator_cd
    from &pam_dolibref..data_sel_content_data_filter
    where data_sel_id eq "&data_selection" and kupcase(content_type_cd) eq "MAINSENSORSOURCE";
  quit;

  proc sort data=work.data_sel_tags_data_filter(where=(data_sel_id eq "&data_selection"))
          out=sortedFilters;
    by tag_id filter_order_no;
  run;
  
  %local dsid;
  %let dsid=%sysfunc(open(work.sortedFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;     /* There is at least one tag filter. */
    /* Create table containing tags which have data value filters. */
    proc sql noprint;
      create table filterTagRKs as
      select distinct put(tag_id,$tagIDToRKFmt.) as &RKColumnName
      from work.sortedFilters;
    quit;
    /* Create table containing tags with no data value filters. */
    proc sql noprint;
      create table tagRKsNoFilter as
      select distinct tagIDsAndRKs.&RKColumnName
      from work.tagIDsAndRKs as allTags
           where allTags.tag_id not in 
		   (select tag_id from work.sortedFilters);
    quit;
  %end;
  %else %do;  /* There are no tags with filters. */
    /* Create empty table since no tags have filters assigned. */
    data filterTagRKs;
      length &RKColumnName 8;
      stop;
    run;
    /* Create table containing tags with no data value filters. */
    proc sql noprint;
      create table tagRKsNoFilter as
      select distinct &RKColumnName
      from work.tagIDsAndRKs;
    quit;
  %end;

  %local dsid;
  %let dsid=%sysfunc(open(work.sortedFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;
    data _null_;
      set work.sortedFilters end=last;
      by tag_id filter_order_no;

      length tempFilter $32767;
      retain tempFilter;

      if first.tag_id then do;
        tempFilter = "( &RKcolumnName eq " || put(tag_id,$tagIDToRKFmt.) || " and ( ";
	    count+1;
      end;
      else do;
        tempFilter = ktrim(tempFilter) || " " || ktrim(kleft(filter_logical_operator_cd));
      end;
	  %if &transposeFlag eq Y %then %do;
        tempFilter = ktrim(tempFilter) || " &measure_val_col_nm " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
	  %else %do;
        tempFilter = ktrim(tempFilter) || " value " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
	  %end;
      if last.tag_id then do;
        tempFilter = ktrim(tempFilter) || " ) ";
        if not last then tempFilter = ktrim(tempFilter) || " ) OR ";
        else tempFilter = ktrim(tempFilter) || " )";
        call symput('phrase'||left(put(count,8.)),ktrim(kleft(tempFilter)));
        call symput('nphrases',left(put(count,8.)));
      end;
    run;

    %let tagFilter=;
    %do i = 1 %to &nphrases;
      %let tagFilter=&tagFilter &&phrase&i;
    %end;
  %end;
%end;

%if &extractEvents eq Y %then %do;
  proc sql noprint;
    create table work.data_sel_events_data_filter as
    select data_sel_id, content_type_cd, content_id as event_id, filter_order_no, filter_value_no, 
           filter_operator_cd, filter_logical_operator_cd
    from &pam_dolibref..data_sel_content_data_filter
    where data_sel_id eq "&data_selection" and kupcase(content_type_cd) eq "MAINEVENTSOURCE";
  run;

  proc sort data=work.data_sel_events_data_filter(where=(data_sel_id eq "&data_selection"))
          out=sortedEventFilters;
    by event_id filter_order_no;
  run;
  
  %local dsid;
  %let dsid=%sysfunc(open(work.sortedEventFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedEventFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;     /* There is at least one tag filter. */
    /* Create table containing events which have data value filters. */
    proc sql noprint;
      create table filterEventRKs as
      select distinct put(event_id,$eventIDToRKFmt.) as &eventRKColumnName 
      from work.sortedEventFilters;
    quit;
    /* Create table containing events with no data value filters. */
    proc sql noprint;
      create table eventRKsNoFilter as
      select distinct &eventRKColumnName 
      from work.eventColumnNames
      where event_id not in (select event_id from work.sortedEventFilters);
    quit;
  %end;
  %else %do; /* There are no event filters. */
    /* Create empty table since no events have filters assigned. */
    data filterEventRKs;
      length &eventRKColumnName 8;
      stop;
    run;
	/* Create table containing events with no data value filters. */
    proc sql noprint;
      create table eventRKsNoFilter as
      select distinct &eventRKColumnName
      from work.eventColumnNames;
    quit;
  %end;


  %local dsid;
  %let dsid=%sysfunc(open(work.sortedEventFilters));
  %if &dsid > 0 %then %do;
    %let nobs =%sysfunc(attrn(&dsid,nobs));
    %let dsid = %sysfunc(close(&dsid));
  %end;
  %else %do;
    %put ERROR: Open for data set work.sortedEventFilters failed.;
    %let pam_gen_fs_rc = 8;
  %end;

  %if &nobs > 0 %then %do;
    data _null_;
      set work.sortedEventFilters end=last;
      by event_id filter_order_no;

      length tempFilter $32767;
      retain tempFilter;

      if first.event_id then do;
        tempFilter = "( &eventRKColumnName eq " || put(event_id,$eventIDtoRKfmt.) || " and ( ";
	    count+1;
      end;
	  else do;
        tempFilter = ktrim(tempFilter) || " " || ktrim(kleft(filter_logical_operator_cd));
      end;
      %if &transposeFlag eq Y %then %do;
        tempFilter = ktrim(tempFilter) || " &event_value_col_nm " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
	  %else %do;
        tempFilter = ktrim(tempFilter) || " value " || ktrim(kleft(filter_operator_cd)) || " " || ktrim(kleft(put(filter_value_no,best.)));
      %end;
      if last.event_id then do;
        tempFilter = ktrim(tempFilter) || " ) ";
        if not last then tempFilter = ktrim(tempFilter) || " ) OR ";
        else tempFilter = ktrim(tempFilter) || " )";
        call symput('phrase'||left(put(count,8.)),kleft(ktrim(tempFilter)));
        call symput('nphrases',left(put(count,8.)));
      end;
    run;

    %let eventFilter=;
    %do i = 1 %to &nphrases;
      %let eventFilter=&eventFilter &&phrase&i;
    %end;
  %end;
%end;

%if %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq Y %then %do;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
    proc sql noprint;
      create table work.tagKeys as
      select distinct pam_tag_dim.&RKColumnName
      from work.tagIDs as tagIDs,
           &pam_dimlibref..&tagDimTable as pam_tag_dim
      where kupcase(tagIDs.tag_id) eq kupcase(pam_tag_dim.tag_id) and 
            tagIDs.tag_id eq pam_tag_dim.&tagIdColumnName and current_num eq 1;
    quit;
  %end;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;
      create table work.eventKeys as
      select distinct pam_event_dim.&eventRKColumnName
      from work.eventIDs as eventIDs,
           &pam_dimlibref..&eventDimTable as pam_event_dim
      where kupcase(eventIDs.event_id) eq kupcase(pam_event_dim.event_id) and
            eventIDs.event_id eq pam_event_dim.&eventIDColumnName and current_num eq 1;
    quit;
  %end;

  %pam_expand_transpose( outds=&pam_fslibref..&filteredSubsetTableName,
                         groupvar=pam_asset_dim_rk pam_loc_dim_rk, 
                         begindttm=&startDatetime,
                         enddttm=&endDatetime, 
                         timeslicebegin=&startTimeSlice,
                         timesliceend=&endTimeSlice, 
                         periodicity=&periodicitySeconds
              %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
			             ,
			             measure_inds=tagData, 
                         measure_dttmvar=&measure_dttm_col_nm, 
                         measure_rkvar=&RKColumnName,  
                         measure_idvar=tag_ID,
                         measure_var=&measure_val_col_nm,
                         measure_colnamevar=&tagTransposeColumnName, 
                         measure_dim=&pam_dimlibref..&tagDimTable,
                         measure_rklist=work.tagKeys, 
                         measure_int=&interpolationMethodTag,
                         measure_agg=&aggregationMethodTag,
                         measure_filter=&tagFilter,
                         measure_filterRKs=work.filterTagRKs, 
                         measure_RKsNoFilter=work.tagRKsNoFilter
			  %end;
              %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
			             ,
                         event_inds=eventData, 
                         event_dttmvar=&event_dttm_col_nm, 
                         event_rkvar=&eventRKColumnName,  
                         event_idvar=event_ID,
                         event_var=&event_value_col_nm,
                         event_colnamevar=&eventTransposeColumnName, 
                         event_dim=&pam_dimlibref..&eventDimTable,
                         event_rklist=work.eventKeys,
                         event_agg=&aggregationMethodEvent,
                         event_filter=&eventFilter,
                         event_filterRKs=work.filterEventRKs, 
                         event_RKsNoFilter=work.eventRKsNoFilter
               %end;
                         );

%end;
%else %if %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq N or
          %sysfunc(ktrim(%sysfunc(kleft(&transposeFlag)))) eq BLANK %then %do;
  %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y and
      %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;	  
      create table tagColumnNames as
      select &RKColumnName, tag_id
      from &pam_dimlibref..&tagDimTable
      where &RKColumnName in (
        %do i=1 %to %eval(&tagRKListLength. -1);
          &&tagRKList&i,
        %end;
          &&tagRKList&tagRKListLength);
    quit;
    data tagRKfmt(keep=start label fmtname type);
      set work.tagColumnNames(rename=(&RKColumnName=start tag_id=label));
      retain fmtname "tagRKfmt" type "N";
    run;
    proc sort data=work.tagRKfmt;
      by start label;
    run;
    proc format cntlin=work.tagRKfmt;
    run;
    proc sql noprint;
      create table work.eventIDs as
      select distinct &eventRKColumnName, event_ID
      from &pam_dimlibref..&eventDimTable
      where &eventRKColumnName in (
         %do i=1 %to %eval(&eventRKListLength. -1);
           &&eventRKList&i,
         %end;
           &&eventRKList&eventRKListLength) 
         and current_num=1;
    quit;
    data eventRKfmt(keep=start label fmtname type);
      set work.eventIDs(rename=(&eventRKColumnName=start event_id=label));
      retain fmtname "eventRKfmt" type "N";
    run;
    proc sort data=work.eventRKfmt;
      by start label;
    run;
    proc format cntlin=work.eventRKfmt;
    run;
    proc sql noprint;
      create table work.assetIDs as
      select distinct pam_asset_dim_rk, pam_asset_ID
      from &pam_dimlibref..&assetDimTable
      where
         %if &assetRKList1 ne %str() %then %do;
           pam_asset_dim_rk in (
              %do i=1 %to %eval(&assetRKListLength. -1);
                &&assetRKList&i,
              %end;
                &&assetRKList&assetRKListLength) and
         %end;
         current_num=1;
    quit;
    data assetRKfmt(keep=start label fmtname type);
      set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label));
      retain fmtname "assetRKfmt" type "N";
    run;
    proc sort data=work.assetRKfmt;
      by start label;
    run;
    proc format cntlin=work.assetRKfmt;
    run;

	%let tagRKsNoFilter1 = ;
    data _null_;
      set work.tagRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numTagRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numTagRKsNoFilter',num);
      end;
    run;

	%let eventRKsNoFilter1 = ;
    data _null_;
      set work.eventRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numEventRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numEventRKsNoFilter',num);
      end;
    run;

    data work.tagData(keep=loc_cd asset_id datetime id value);
      set work.tagData (rename=(&measure_dttm_col_nm=DATETIME &measure_val_col_nm=VALUE)
                        where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        ));
      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ID = put(&RKColumnName,tagRKfmt.);
	  
      %if &tagFilter ne %str() %then %do;
        if &tagFilter
          %if &tagRKsNoFilter1 ne %str() %then %do;
             or (&RKColumnName in (
               %do i = 1 %to &numTagRKsNoFilter;
                 &&tagRKsNoFilter&i
               %end;
             ))
          %end;
        then output;
      %end;

    run;
    data work.eventData(keep=loc_cd asset_id datetime id value);
      set work.eventdata (rename=(&event_dttm_col_nm=DATETIME &event_value_col_nm=VALUE)
                          where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        )) end=eof;
      ID = put(&eventRKColumnName,eventRKfmt.);
	  
      %if &eventFilter ne %str() %then %do;
        if &eventFilter
          %if &eventRKsNoFilter1 ne %str() %then %do;
             or (&eventRKColumnName in (
               %do i = 1 %to &numEventRKsNoFilter;
                 &&eventRKsNoFilter&i
               %end;
             ))
           %end;
        then output;
      %end;

    run;
    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id datetime id value);
      set work.tagData work.eventData end=eof;
      numObs + 1;
      if (eof) then call symputx("pam_fs_num_obs", numObs);
    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;
  %end;
  %else %if %sysfunc(ktrim(%sysfunc(kleft(&extractMeasures)))) eq Y %then %do;
    proc sql noprint;
	  create table tagColumnNames as
	  select &RKColumnName, tag_id
	  from &pam_dimlibref..&tagDimTable
	  where &RKColumnName in (
         %do i=1 %to %eval(&tagListLength. -1);
           &&tagRKList&i,
         %end;
           &&tagRKList&tagRKListLength);
	quit;
    data tagRKfmt(keep=start label fmtname type);
      set work.tagColumnNames(rename=(&RKColumnName=start tag_id=label));
      retain fmtname "tagRKfmt" type "N";
    run;
    proc sort data=work.tagRKfmt;
      by start label;
    run;
    proc format cntlin=work.tagRKfmt;
    run;

    data _null_;
      set work.tagRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&RKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numTagRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('tagRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numTagRKsNoFilter',num);
      end;
    run;

    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id datetime id value);
      set work.tagData (rename=(&measure_dttm_col_nm=DATETIME &measure_val_col_nm=VALUE)
                        where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        )) end=eof;

      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ID = put(&RKColumnName,tagRKfmt.);

      %if &tagFilter ne %str() %then %do;
        if &tagFilter
          %if &tagRKsNoFilter1 ne %str() %then %do;
             or (&RKColumnName in (
               %do i = 1 %to &numTagRKsNoFilter;
                 &&tagRKsNoFilter&i
               %end;
             ))
          %end;

        then do; 
          numObs + 1;
          output;
        end;
      %end;
      %else %do;
        numObs + 1;
      %end;
      
      if (eof) then call symputx("pam_fs_num_obs", numObs);

    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;

  %end;
  %else %if %sysfunc(ktrim(%sysfunc(kleft(&extractEvents)))) eq Y %then %do;
    proc sql noprint;
      create table work.eventIDs as
      select distinct &eventRKColumnName, event_ID
      from &pam_dimlibref..&eventDimTable
      where &eventRKColumnName in (
         %do i=1 %to %eval(&eventRKListLength. -1);
           &&eventRKList&i,
         %end;
           &&eventRKList&eventRKListLength) 
         and current_num=1;
    quit;
    data eventRKfmt(keep=start label fmtname type);
      set work.eventIDs(rename=(&eventRKColumnName=start event_id=label));
      retain fmtname "eventRKfmt" type "N";
    run;
    proc sort data=work.eventRKfmt;
      by start label;
    run;
    proc format cntlin=work.eventRKfmt;
    run;
    proc sql noprint;
      create table work.assetIDs as
      select distinct pam_asset_dim_rk, pam_asset_ID
      from &pam_dimlibref..&assetDimTable
      where pam_asset_dim_rk in (
         %do i=1 %to %eval(&assetRKListLength. -1);
           &&assetRKList&i,
         %end;
           &&assetRKList&assetRKListLength) 
         and current_num=1;
    quit;
    data assetRKfmt(keep=start label fmtname type);
      set work.assetIDs(rename=(pam_asset_dim_rk=start pam_asset_id=label));
      retain fmtname "assetRKfmt" type "N";
    run;
    proc sort data=work.assetRKfmt;
      by start label;
    run;
    proc format cntlin=work.assetRKfmt;
    run;

	%let eventRKsNoFilter1 = ;
    data _null_;
      set work.eventRKsNoFilter;
      length tempList $32767;
      retain tempList;
      retain num 1;
      if tempList eq "" or tempList eq "," then do;
        tempList = ktrim(kleft(tempList))||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      else do;
        tempList=ktrim(kleft(tempList))||","||ktrim(kleft(put(&eventRKColumnName,8.)));
      end;
      if mod(_n_,327) eq 0 then do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),tempList);
        call symput('numEventRKsNoFilter',left(put(num,8.)));
        tempList=",";
        num+1;
      end;
      else do;
        call symput('eventRKsNoFilter'||left(put(num,8.)),ktrim(kleft(tempList)));
        call symputx('numEventRKsNoFilter',num);
      end;
    run;

    data &pam_fslibref..&filteredSubsetTableName(keep=loc_cd asset_id id datetime value);
      set work.eventData (rename=(&event_dttm_col_nm=DATETIME &event_value_col_nm=VALUE)
                          where = (datetime between "&startDatetime"dt and "&endDatetime"dt

      %if &startTimeSlice ne %str() and &endTimeSlice ne %str() %then %do;
        and timepart(datetime) between "&startTimeSlice"t and "&endTimeSlice"t
      %end;
        )) end=eof;

      length ID $100;
      attrib DATETIME label="DATETIME";
      attrib VALUE label = "VALUE";

      ASSET_ID = put(pam_asset_dim_rk,assetRKfmt.);
      ID = put(&eventRKColumnName,eventRKfmt.);

      %if &eventFilter ne %str() %then %do;
        if &eventFilter
          %if &eventRKsNoFilter1 ne %str() %then %do;
             or (&eventRKColumnName in (
               %do i = 1 %to &numEventRKsNoFilter;
                 &&eventRKsNoFilter&i
               %end;
             ))
          %end;

        then do;
          numObs + 1;
          output;
        end;
      %end;
      %else %do;
        numObs + 1;
      %end;
	  
      if (eof) then call symputx("pam_fs_num_obs", numObs);

    run;
    proc sort data=&pam_fslibref..&filteredSubsetTableName;
      by loc_cd asset_id datetime;
    run;
  %end;
%end;
%mend;

%macro pam_gen_fs_get_fs_attrs;
/* Begin the query. The DATA statement names the output data set.  */
data work.metadata_libraries;

  /* The LENGTH statement defines variables for function arguments and */
  /* assigns the maximum length of each variable.                      */
  length liburi upasnuri $256 name $128 type id $17 libref engine $8 path mdschemaname schema $256;

  /* The KEEP statement defines the variables to include in the output data set.  */
  keep name libref engine path;

  /* The CALL MISSING routine initializes the output variables to missing values.  */
  call missing(liburi,upasnuri,name,engine,libref);

  /* The METADATA_GETNOBJ function specifies to get the SASLibrary objects            */
  /* in the repository. The argument nlibobj=1 specifies to get the first object that */
  /* matches the requested URI. liburi is an output variable. It will store the URI   */
  /* of the returned SASLibrary object.                                               */
  nlibobj=1;
  librc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",nlibobj,liburi);

  /* Initialize found flag.  Flag indicates when PAMFS library information has been found. */
  found = 0;

  /* The DO statement specifies a group of statements to be executed as a unit           */
  /* for each object that is returned by METADATA_GETNOBJ. The METADATA_GETATTR function */
  /* is used to retrieve the values of the Name, Engine, and Libref attributes of        */
  /* the SASLibrary object.                                                              */

  do while (librc>0 and found=0);
    /* Get Library attributes */
    rc=metadata_getattr(liburi,'Name',name);
    rc=metadata_getattr(liburi,'Engine',engine);
    rc=metadata_getattr(liburi,'Libref',libref);

    if kupcase(libref) eq kupcase("&pam_fslibref") then do;
 
      /* The PAMFS library has been found.  Set flag to exit loop after current iteration. */
      found = 1;

      /* The METADATA_GETNASN function specifies to get objects associated to the           */
      /* library via the UsingPackages association. The n argument specifies to return the  */
      /* first associated object for that association type. upasnuri is an output variable. */
      /* It will store the URI of the associated metadata object, if one is found.          */
      n=1;
      uprc=metadata_getnasn(liburi,'UsingPackages',n,upasnuri);

      /* When a UsingPackages association is found, the METADATA_RESOLVE function           */
      /* is called to resolve the URI to an object on the metadata server. The CALL MISSING */
      /* routine assigns missing values to output variables.                                */
      if uprc > 0 then do;
        call missing(type,id,path,mdschemaname,schema);
        rc=metadata_resolve(upasnuri,type,id);

        /* If type='Directory', the METADATA_GETATTR function is used to get its */
        /* path and output the record                                            */
        if type='Directory' then do;
          rc=metadata_getattr(upasnuri,'DirectoryName',path);
          if kupcase(libref) eq kupcase("&pam_fslibref") then do;
            output;
          end;
        end; 

        /* If type='DatabaseSchema', the METADATA_GETATTR function is used to get */ 
        /* the name and schema, and output the record                             */
        else if type='DatabaseSchema' then do;
          rc=metadata_getattr(upasnuri,'Name',mdschemaname);
          rc=metadata_getattr(upasnuri,'SchemaName',schema);
          if kupcase(libref) eq kupcase("&pam_fslibref") then do;
            output;
          end;
        end; 

        /* Check to see if there are any more Directory objects */
        n+1;
        uprc=metadata_getnasn(liburi,'UsingPackages',n,upasnuri);
      end; /* if uprc > 0 */
    end; /* if libref = pamfs */

    /* Look for another library */
    nlibobj+1;
    librc=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",nlibobj,liburi);
  end; /* do while (librc>0) and found = 0 */
run;
%if &sqlobs eq 0 %then %do;
  /* PAMFS library information was not found in the metadata server. */
  %let pam_fs_size = .;
%end;
%else %do;
  proc sql noprint;
    select distinct path into :fsPath
    from work.metadata_libraries
    where kupcase(libref) eq kupcase("&pam_fslibref");
  quit;
  
  data _null_;
    %if &sysscp eq WIN %then %do;
      file='"'||ktrim(kleft(%quote(tranwrd("&fsPath","/","\"))))||%str("\")||ktrim(kleft("&data_sel_id"))||".sas7bdat"||'"';
    %end;
    %else %if &sysscp eq LINUX or &sysscp eq %str(LIN X64) %then %do;
      file='"'||ktrim(kleft(%quote(tranwrd("&fsPath","\","/"))))||%str("/")||klowcase(ktrim(kleft("&data_sel_id")))||".sas7bdat"||'"';
    %end;

    rc=filename("physFile",%quote(file));
    fid=fopen("physFile");
    %if &sysscp eq WIN %then %do;
      filesizeKB=input(finfo(fid,foptname(fid,4)),32.)/1024;
    %end;
    %else %if &sysscp eq LINUX or &sysscp eq %str(LIN X64) %then %do;
      filesizeKB=input(finfo(fid,foptname(fid,6)),32.)/1024;
    %end;

    call symputx('pam_fs_size', fileSizeKB);
    close=fclose(fid);
  run;

%end;  /* End if PAMFS library was found in metadata */
  
%if %sysfunc(exist(&pam_fslibref..&data_sel_id)) %then %do;
  data _null_;
    dsid = open("&pam_fslibref..&data_sel_id");
    if (dsid eq 0) then do; 
      return;
    end;
    else do;
      call symput('pam_fs_num_cols', put(attrn(dsid, "NVARS"),8.));
      %if &pam_fs_num_obs eq . %then %do;
        call symputx('pam_fs_num_obs', attrn(dsid, "NOBS"));
      %end;
    end;
    dsid = close(dsid);
  run;
%end;

%finish:

%mend;

%macro pam_gen_fs(data_selection_id=);
 /*+----------------------------------------------------------------------+
   |
   | Name:     pam_gen_fs macro
   |
   | Purpose:  This is the main section of the Data Selection macro
   |           code.  There are several macros involved in the
   |           creation of a filtered subset.
   |
   |
   | Product:  SAS Predictive Asset Maintenance
   | Support:  Cheryl Davis
   |
   | Parms:    Data Selection ID
   |
   | Output:   Filtered Subset
   |
   | History:  date    action
   |           30Sep11 Initial coding                               cld
   |           23Aug12 Support events data - untransposed           cld
   |           04Sep12 Support events data - transposed             cld
   |           04Sep12 Integrate with major changes in 
   |                   pam_expand_transpose macro for 5.2           cld 
   |           04Sep12 Read mart metadata from new pamdo tables in
   |                   Manufacturing Commons structure              cld
   |           01Oct12 Add support for filtering of events data     cld
   |           16Oct12 Remove dictionary table references           cld
   |           03Apr13 Count number of obs in filtered subset 
   |                   as they are written                          cld
   |
   | Notes:    None
   |
   | Copyright (c) 2012 by SAS Institute Inc., Cary, NC, USA.
   | All Rights Reserved.
   |
   +----------------------------------------------------------------------+*/

%pam_prebuild;

/* Capture current options settings to put them back at the */
/* end of the data selection.                               */
%let prev_mprint = %sysfunc(getoption(mprint));
%let prev_mlogic = %sysfunc(getoption(mlogic));
%let prev_symbolgen = %sysfunc(getoption(symbolgen));
%let prev_notes = %sysfunc(getoption(notes));
%let prev_source = %sysfunc(getoption(source));

/* Options statement for data selection debugging options. */ 
/* These macro variables are set in pam_prebuild.sas.      */
options &ds_mprint &ds_mlogic &ds_symbolgen &ds_notes &ds_source;

%global filteredSubsetTableName;
%global measure_val_col_nm;
%global measure_dttm_col_nm;
%global event_value_col_nm;
%global event_dttm_col_nm;
%global tagDimTable;
%global eventDimTable;
%global assetDimTable;
%global locDimTable;
%global tagID;
%global minval;
%global maxval;
%global lowerBoundOperator;
%global upperBoundOperator;
%global logicalOperator;
%global interpolationMethod;
%global interpolationInterval;
%global tagRKList;
%global eventRKList;
%global assetRKList;
%global p_table_names;
%global RKcolumnName;
%global eventRKcolumnName;
%global tagIdColumnName;
%global eventIdColumnName;
%global tagTransposeColumnName;
%global eventTransposeColumnName;
%global transposeFlag;
%global startDate;
%global endDate;
%global startTimeSlice;
%global endTimeSlice;
%global startDatetime;
%global endDatetime;
%global pam_fs_num_cols;
%global pam_fs_num_obs;
%global pam_fs_size;
%global pam_gen_fs_rc;
%global extractEvents;
%global extractMeasures;
%global interpolationMethodTag;
%global interpolationMethodEvent;
%global aggregationMethodTag;
%global aggregationMethodEvent;
%global periodicitySeconds;
%global contentTypeCodes;

%local measureFactTable;
%local eventFactTable;

%let startDate=;
%let endDate=;
%let startTimeSlice=;
%let endTimeSlice=;
%let startDatetime=;
%let endDatetime=;
%let filteredSubsetTableName=;
%let measure_val_col_nm=;
%let measure_dttm_col_nm=;
%let event_value_col_nm=;
%let event_dttm_col_nm=;
%let tagDimTable=pam_tag_dim;
%let eventDimTable=pam_event_dim;
%let assetDimTable=pam_asset_dim;
%let locDimTable=pam_loc_dim;
%let tagID=;
%let minval=;
%let maxval=;
%let lowerBoundOperator=;
%let upperBoundOperator=;
%let logicalOperator=;
%let interpolationMethod=;
%let interpolationInterval=;
%let tagRKList=;
%let eventRKList=;
%let assetRKList=;
%let p_table_names=;
%let RKcolumnName=;
%let eventRKcolumnName=;
%let tagIdColumnName=;
%let eventIdColumnName=;
%let tagTransposeColumnName=;
%let eventTransposeColumnName=;
%let transposeFlag=;
%let pam_fs_num_cols=;
%let pam_fs_num_obs = .;
%let pam_fs_size=;
%let pam_gen_fs_rc = 0;
%let extractEvents = N;
%let extractMeasures = N;
%let interpolationMethodTag=;
%let interpolationMethodEvent=;
%let aggregationMethodTag=;
%let aggregationMethodEvent=;
%let periodicitySeconds=3600;

data &pam_dolibref..data_sel;
  modify &pam_dolibref..data_sel;
  if data_sel_id eq kupcase("&data_selection_id") or data_sel_id eq kupcase("&p_data_selection_id");
  run_dttm = datetime();
  data_sel_status_cd = "RUNNING";
run;

%if &p_data_selection_id ne "" %then %let data_sel_id = %sysfunc(kupcase(&p_data_selection_id));
%else %if &data_selection_id ne "" %then %let data_sel_id = %sysfunc(kupcase(&data_selection_id));
%else %do;
  %put ERROR: Data selection ID is required.;
  %let pam_gen_fs_rc = 8;
  %goto finish;
%end;

proc sql noprint;
  select data_sel_id
  from &pam_dolibref..data_sel
  where data_sel_id eq "&data_sel_id";
quit;

%if &sqlobs eq 0 %then %do;
  %put ERROR: No data selection definition exists for selection ID &data_sel_id.;
  %let pam_gen_fs_rc = 8;
  %goto finish;
%end;

data _null_;
  set &pam_dolibref..data_sel_fact;
  if data_sel_id eq "&data_sel_id";

  if upcase(fact_tbl_cd) eq "TAGS" then do;
    call symput('extractMeasures',"Y");
  end;
  if upcase(fact_tbl_cd) eq "EVENTS" then do;
    call symput('extractEvents',"Y");
  end;
run;

%pam_gen_fs_get_dim_table_names;

proc sql noprint;
  create table work.locations as
  select distinct loc_cd, pam_loc_dim_rk
  from &pam_dimlibref..&locDimTable as pam_loc_dim
  where current_num eq 1;
quit;
data locfmt(keep=fmtname start label type);
  set work.locations;
  retain fmtname "locfmt" type "N";
  start=pam_loc_dim_rk;
  label=loc_cd;
run;
proc sort data=work.locfmt;
  by start label;
run;
proc format cntlin=locfmt;
run;

%if &extractMeasures eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_mart_meta_tag;
%end;
%if &extractEvents eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_mart_meta_event;
%end;

%pam_gen_fs_get_data_sel(data_selection=&data_sel_id);

data _null_;
  set &pam_dolibref..data_sel_content(where=(data_sel_id eq "&data_sel_id"));
  length temp $100;
  retain temp;
  retain foundTags foundEvents 0;

  if not foundTags and kupcase(content_type_cd) eq "MAINSENSORSOURCE" then do;
    temp=ktrim(kleft(temp))||" TAGS";
    foundTags = 1;
  end;
  else if not foundEvents and kupcase(content_type_cd) eq "MAINEVENTSOURCE" then do;
    temp=ktrim(kleft(temp))||" EVENTS";
    foundEvents = 1;
  end;
  call symput('contentTypeCodes',temp);
run;

%pam_gen_fs_get_content(contentTypes=&contentTypeCodes, dataSelection=&data_sel_id);
%pam_gen_fs_get_filter_content(filterContentTypes=&contentTypeCodes, dataSelection=&data_sel_id);

%if &extractMeasures eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_tags(data_selection=&data_sel_id);
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_tag_RKs;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_extra_tag_obs;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_measure_facts;
%end;

%if &extractEvents eq Y %then %do;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_assets(data_selection=&data_sel_id);
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_asset_RKs;
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_list_of_events(data_selection=&data_sel_id);
  %if &pam_gen_fs_rc le 4 %then %pam_gen_fs_get_event_facts;
%end;

%if &pam_gen_fs_rc le 4 %then %pam_gen_fs_transpose(data_selection=&data_sel_id);

%finish:

%pam_gen_fs_get_fs_attrs;

/* Update PAMDO.DATA_SEL table with filtered subset attributes. */
data &pam_dolibref..data_sel;
  modify &pam_dolibref..data_sel;
  if data_sel_id eq "&data_sel_id";
  %if &pam_gen_fs_rc le 7 %then %do;
    filtered_subset_rows_no = &pam_fs_num_obs;
    filtered_subset_columns_no = &pam_fs_num_cols;
    filtered_subset_size_no = &pam_fs_size;
    data_sel_status_cd = "SUCCESS";
  %end;
  %else %do;
    data_sel_status_cd = "FAILURE";
  %end;
run;

%if &transposeFlag eq BLANK %then %do;
  %put **** WARNING **** WARNING **** WARNING **** WARNING **** WARNING **** WARNING **** ;
  %put;
  %put WARNING: DATA_SEL.DATA_SEL_TRANSPOSE_FLG contains a missing value for ;
  %put %str(        ) this data selection. ;
  %put;
  %put %str(        ) The resulting filtered subset %sysfunc(upcase(&pam_fslibref)).&data_sel_id will contain ;
  %put %str(        ) non-transposed data, regardless of the selection that was ;
  %put %str(        ) made for the transpose option in the Data Selection Wizard. ;
  %put;
  %put ***********************************************************************************;
  %put;   %put;   %put;
%end;

%put ******************** PAM Data Selection Report *********************;
%put *                                                           ;
%put * %str(  )Filtered Subset: %sysfunc(upcase(&pam_fslibref)).&data_sel_id ;
%put *                                                           ;
%if &pam_gen_fs_rc gt 7 %then %do;
  %put * %str(  )Data selection status:          Failure         ;
%end;
%else %if &pam_gen_fs_rc ge 4 %then %do;
  %put * %str(  )Data selection status:          Completed with Warnings ;
  %put * %str(  )Number of variables:           &pam_fs_num_cols ;
  %put * %str(  )Number of observations:        &pam_fs_num_obs  ;
  %put * %str(  )Size of filtered subset (KB):  &pam_fs_size     ;
%end;
%else %do;
  %put * %str(  )Data selection status:          Success         ;
  %put * %str(  )Number of variables:           &pam_fs_num_cols ;
  %put * %str(  )Number of observations:        &pam_fs_num_obs  ;
  %put * %str(  )Size of filtered subset (KB):  &pam_fs_size     ;
%end;
%put *                                                           ;
%put *********************************************************************;

/* Put debugging options back to their previous settings. */
options &prev_mprint &prev_mlogic &prev_notes &prev_source &prev_symbolgen;
%mend;
