/******************************************************************
* 
* Name:  pam_get_model_elements_from_BRM
* 
* Arguments: 
*
* BRMLIB = WORK: DO NOT CHANGE! location where brs_decision_flow macro puts its temporary datasets
* OUTLIB: location where user wants to put the data sets needed for Model Manager
* rulesetin = rulesetinput: DO NOT CHANGE! The name of the temporary data set created by brs_decision_flow
* newinput: The name of the input data set for Model Manager, e.g. modelinput
* rulesetout = rulesetoutput: DO NOT CHANGE! The name of the temporary data set created by brs_decision_flow
* newoutput:  The name of the output data set for Model Manager, e.g. modeloutput
* outfilepath:  location of the score code, called logic-only.sas 
* SAS_CONFIG:  location of the SAS Config directory, e.g. C:\SAS\Config
* METASERVER: connection info for metadata server, 
* METAPORT: connection info for metadata server, 
* METAUSER: connection info for metadata server,
* METAPASS: connection info for metadata server,
* publishedflowname: name of the BRM published rule flow, e.g. MyFlow(10060)
* brmversion:  version number of the BRM published rule flow, e.g  1.2
*
*
*
*******************************************************************/
%macro TransposeRulesDS(
	ds_in = work.Rulesetinput,
	ds_out = work.ds_to_score
);
proc sql noprint;
	create table ds_distinct as select distinct TermId, dataType from &ds_in;
quit;
data ds_in_num ds_in_char;
	set ds_distinct;
	length sampleChar $128;
	if upcase(trim(left(dataType)))="STRING" then do;
		sampleChar = "Yes";
		output ds_in_char;
	end;
	else do;
		sampleNum = 5;
		output ds_in_num;
	end;
run;
proc transpose data=ds_in_num out=out_num(drop=_NAME_);
	id TermId;
	var sampleNum;
run;
proc transpose data=ds_in_char out=out_char(drop=_NAME_);
	id TermId;
	var sampleChar;
run;
data &ds_out;
	merge out_num out_char;
run;
%mend;

%macro pam_get_model_elements_from_BRM(
	BRMLIB = WORK,
	OUTLIB = WORK,
	rulesetin = rulesetinput,
	newinput = modelinput,
	rulesetout = rulesetoutput,
	newoutput = modeloutput,
	outfilepath = C:\,
	SAS_CONFIG = C:\SAS\Config,
	METASERVER=, 
	METAPORT=, 
	METAUSER=,
	METAPASS=,
	publishedflowname=,
	brmversion=
);


%let PUBLISHED_FLOW_NAME=&publishedflowname;
%let VERSION=&brmversion; 


%SVCREG(NAME=BrmPublishedBusinessRuleFlow REST 
, TYPE=REST 
, METASERVER=&METASERVER 
, METAPORT=&METAPORT 
, METAUSER=&METAUSER 
, METAPASS=&METAPASS ); 

%put "&SERVICE_ENDPOINT";

filename testout "&outfilepath.prochttp.txt";

proc http out=testout
          url="&SERVICE_ENDPOINT/&publishedflowname/&version"
          method="GET"   
;
run;

%include "&SAS_CONFIG/Lev1/Applications/SASBusinessRulesManager/1.2/ucmacros/brs_decision_flow.sas";

%brs_decision_flow(InputTable=,Mapping=,
						 Filelocation=testout,
                         Debugcodeloc=,
                         logicCodeLoc= &outfilepath.logic-only.sas,
                         RuleFire=N);

%TransposeRulesDS(
	ds_in = &BRMLIB..&rulesetin , /* work.Rulesetinput */
	ds_out = &OUTLIB..&newinput   /* the ds_to_score */
);

%TransposeRulesDS(
	ds_in = &BRMlIB..&rulesetout, /* work.Rulesetoutput */
	ds_out = &OUTLIB..&newoutput  /* the scored data set */
);
%mend;


/************************************************************************************/

/* Sample Usage */
	
/*
%pam_get_model_elements_from_BRM(
	BRMLIB = WORK,
	OUTLIB = WORK,
	rulesetin = rulesetinput,
	newinput = modelinput,
	rulesetout = rulesetoutput,
	newoutput = modeloutput,
	outfilepath = C:\ModelLifeCycle\BRM\,
	SAS_CONFIG = C:\SAS\Config,
	METASERVER=abc123.na.sas.com, 
	METAPORT=8561, 
	METAUSER=sadm@saspw,
	METAPASS=Secret123,
	publishedflowname=ApprovalFlow(10000),
	brmversion=1.3
);
*/

/*
%pam_get_model_elements_from_BRM(
	BRMLIB = WORK,
	OUTLIB = WORK,
	rulesetin = rulesetinput,
	newinput = modelinput,
	rulesetout = rulesetoutput,
	newoutput = modeloutput,
	outfilepath = C:\ModelLifeCycle\BRM\,  
	SAS_CONFIG = C:\SAS\Config,
	METASERVER=abc123.na.sas.com, 
	METAPORT=8561, 
	METAUSER=sasadm@saspw,
	METAPASS=Secret123,
	publishedflowname=PamExample2Flow2(10060),
	brmversion=1.2
);
*/
/************************************************************************************/

				
